import textwrap

from .base import RegisteredCommandBase, RegisteredSubCommandBase
from ..parsers import object_id_parser


# Command registry - don't touch!
registry = []


###############################################################################
# Base command
###############################################################################
class DeleteCommand(RegisteredCommandBase):
    name = "delete"
    description = "Delete a superevent signoff"
    subcommands = registry


###############################################################################
# Subcommands - registered to base command automatically
###############################################################################
class DeleteChildBase(RegisteredSubCommandBase):
    _registry = registry


class DeleteSignoffCommand(DeleteChildBase):
    name = "signoff"
    description = "Delete a superevent signoff"
    long_description = textwrap.dedent("""\
        Delete an operator or advocate signoff. Only allowed for superevents;
        event signoff deletion is not presently implemented.
    """).rstrip()
    parent_parsers = (object_id_parser,)

    def add_custom_arguments(self, parser):
        parser.add_argument(
            "signoff_type",
            type=str,
            help=("Signoff type (do '{prog} info signoff_types') to see "
                  "options").format(prog=self.base_prog)
        )
        parser.add_argument(
            "instrument",
            type=str,
            nargs='?',
            help=("Instrument code (do {prog} info instruments to see "
                  "options). Required for operator signoffs.")
            .format(prog=self.base_prog)
        )
        return parser

    def run(self, client, args):
        instrument = args.instrument or ''  # Convert None to ''
        return client.delete_signoff(args.object_id, args.signoff_type,
                                     instrument=instrument)



import pytest
from ligo.gracedb.exceptions import HTTPError

# Apply module-level mark
pytestmark = pytest.mark.integration


def test_basic_creation_and_retrieval(client, create_event):
    event = create_event()
    response = client.createSuperevent(1, 2, 3, event['graceid'], category='T')
    assert response.status_code == 201
    data = response.json()
    assert data['preferred_event'] == event['graceid']
    assert data['t_start'] == 1
    assert data['t_0'] == 2
    assert data['t_end'] == 3
    assert data['category'] == 'Test'

    response = client.superevent(data['superevent_id'])
    assert response.status_code == 200
    data = response.json()
    assert data['preferred_event'] == event['graceid']
    assert data['t_start'] == 1
    assert data['t_0'] == 2
    assert data['t_end'] == 3
    assert data['category'] == 'Test'


def test_creation_with_already_used_preferred_event(client, create_event):
    event = create_event()
    response = client.createSuperevent(1, 2, 3, event['graceid'], category='T')
    assert response.status_code == 201
    data = response.json()
    assert data['preferred_event'] == event['graceid']
    assert data['t_start'] == 1
    assert data['t_0'] == 2
    assert data['t_end'] == 3
    assert data['category'] == 'Test'

    # Try to create new superevent with an existing
    # preferred event:
    with pytest.raises(HTTPError):
        response = client.createSuperevent(1, 2, 3, event, category='T')


def test_creation_with_events(client, create_event):
    event1 = create_event()
    event2 = create_event()
    event3 = create_event()
    response = client.createSuperevent(
        1, 2, 3, event1['graceid'], category='T',
        events=[event2['graceid'], event3['graceid']]
    )
    assert response.status_code == 201
    data = response.json()
    assert data['preferred_event'] == event1['graceid']
    assert data['t_start'] == 1
    assert data['t_0'] == 2
    assert data['t_end'] == 3
    assert data['category'] == 'Test'
    assert len(data['gw_events']) == 3
    for ev in [event1, event2, event3]:
        assert ev['graceid'] in data['gw_events']


def test_creation_with_labels(client, create_event):
    labels = ['INJ', 'DQV']
    event = create_event()
    response = client.createSuperevent(1, 2, 3, event['graceid'], category='T',
                                       labels=labels)
    assert response.status_code == 201
    data = response.json()
    assert data['preferred_event'] == event['graceid']
    assert data['t_start'] == 1
    assert data['t_0'] == 2
    assert data['t_end'] == 3
    assert data['category'] == 'Test'
    assert len(data['labels']) == len(labels)
    for l in labels:
        assert l in data['labels']


@pytest.mark.parametrize(
    "params",
    [
        {'t_0': 3463},
        {'t_start': 208, 't_end': 28992},
        {'preferred_event': True},
        {'t_start': 972, 't_0': 1028, 't_end': 1098, 'preferred_event': True},
        {'t_start': 972, 't_0': 1028, 't_end': 1098, 'preferred_event': True},
        {'t_start': 1024, 't_0': 2048, 't_end': 4096, 'preferred_event': True,
         'em_type': 'EM_TYPE_TEST'},
        {'t_start': 1025, 't_0': 2049, 't_end': 4097, 'preferred_event': True,
         'em_type': 'EM_TYPE_TEST', 'time_coinc_far': 1234.0},
        {'t_start': 1025, 't_0': 2049, 't_end': 4097, 'preferred_event': True,
         'em_type': 'EM_TYPE_TEST', 'time_coinc_far': 1234.0,
         'space_coinc_far': 12345.0},
    ]
)
def test_update(client, create_event, create_superevent, params):
    # Create a superevent
    superevent = create_superevent()

    # Create a new event if we're going to change the preferred_event
    if 'preferred_event' in params:
        event = create_event()
        params['preferred_event'] = event['graceid']

    # Update the superevent
    response = client.updateSuperevent(superevent['superevent_id'], **params)

    # Check results
    assert response.status_code == 200
    data = response.json()
    for p in params:
        assert data[p] == params[p]

    # Check that old preferred event is still in the superevent's events
    if 'preferred_event' in params:
        assert len(data['gw_events']) == 2
        assert event['graceid'] in data['gw_events']


def test_event_addition_and_removal(client, create_event, create_superevent):
    # Create superevent and another event
    superevent = create_superevent()
    new_event = create_event()

    # Add new event to superevent
    response = client.addEventToSuperevent(superevent['superevent_id'],
                                           new_event['graceid'])

    # Check results
    assert response.status_code == 201
    data = response.json()
    assert data['graceid'] == new_event['graceid']

    # Pull down superevent and check results
    response = client.superevent(superevent['superevent_id'])
    assert response.status_code == 200
    data = response.json()
    # Preferred event didn't change
    assert data['preferred_event'] == superevent['preferred_event']
    # Both events are in list
    assert len(data['gw_events']) == 2
    assert superevent['preferred_event'] in data['gw_events']
    assert new_event['graceid'] in data['gw_events']

    # Remove event
    response = client.removeEventFromSuperevent(superevent['superevent_id'],
                                                new_event['graceid'])
    assert response.status_code == 204

    # Pull down superevent and check results
    response = client.superevent(superevent['superevent_id'])
    assert response.status_code == 200
    data = response.json()
    # Preferred event didn't change
    assert data['preferred_event'] == superevent['preferred_event']
    # Only one event in list
    assert len(data['gw_events']) == 1
    assert superevent['preferred_event'] in data['gw_events']
    assert new_event['graceid'] not in data['gw_events']

    # Try remove event again, should get error
    with pytest.raises(HTTPError):
        client.removeEventFromSuperevent(
            superevent['superevent_id'], new_event['graceid']
        )

    # Try remove preferred event, should get error
    with pytest.raises(HTTPError):
        client.removeEventFromSuperevent(
            superevent['superevent_id'], superevent['preferred_event']
        )


def test_confirm_as_gw(client, create_superevent):
    # Create superevent
    superevent = create_superevent()

    # Check its status
    assert superevent['gw_id'] is None
    assert not superevent['superevent_id'].startswith('TGW')
    assert 'GW' not in superevent['superevent_id'][:3]

    # Confirm it as a GW
    response = client.confirm_superevent_as_gw(superevent['superevent_id'])
    assert response.status_code == 200
    data = response.json()
    gw_id = data['gw_id']
    assert gw_id.startswith('TGW') or 'GW' in gw_id[:3]

    # Try to update again
    with pytest.raises(HTTPError):
        client.confirm_superevent_as_gw(superevent['superevent_id'])

    # Test getting data using both IDs
    response = client.superevent(superevent['superevent_id'])
    assert response.status_code == 200
    s_id_data = response.json()
    response = client.superevent(gw_id)
    assert response.status_code == 200
    gw_id_data = response.json()
    for k in s_id_data:
        assert s_id_data[k] == gw_id_data[k]


def test_search(client, create_superevent):
    # Create a superevent
    superevent = create_superevent()

    # Search by ID
    response = client.superevents(superevent['superevent_id'])
    results_list = list(response)
    assert len(results_list) == 1
    assert results_list[0]['superevent_id'] == superevent['superevent_id']



# Tests for subcommands below 'create'
#  Ex: 'gracedb create emobservation'
import pytest
import shlex
try:
    from unittest import mock
except ImportError:
    import mock

# Apply module-level mark
pytestmark = pytest.mark.cli


###############################################################################
# Tests of individual subcommands #############################################
###############################################################################
def test_update_event_subcommand(CLI):
    """Test update event subcommand"""
    cmd_args = {
        'g_id': 'G1234',
        'filename': '/path/to/fake/file.xml',
    }

    # Generate command
    cmd = 'update event {g_id} {filename}'.format(**cmd_args)

    func = 'ligo.gracedb.rest.GraceDb.replaceEvent'
    with mock.patch(func) as mock_cli_func:
        CLI(shlex.split(cmd))

    # Check call count
    assert mock_cli_func.call_count == 1

    # Get args used in function call
    cli_args, cli_kwargs = mock_cli_func.call_args

    # Check args used in function call
    assert cli_args == (cmd_args['g_id'], cmd_args['filename'],)
    assert cli_kwargs == {}


SIGNOFF_TEST_DATA = [
    (None, None, None),
    ('FAKE_INSTRUMENT', 'FAKE_STATUS', 'test comment'),
]
@pytest.mark.parametrize("instrument,status,comment",  # noqa: E302
                         SIGNOFF_TEST_DATA)
def test_update_signoff_subcommand(CLI, instrument, status, comment):
    """Test update signoff subcommand"""
    cmd_args = {
        's_id': 'S001122a',
        'signoff_type': 'FAKE_SIGNOFF_TYPE',
    }

    # Generate command
    cmd = "update signoff {s_id} {signoff_type}".format(**cmd_args)
    if instrument is not None:
        cmd += " {inst}".format(inst=instrument)
    if status is not None:
        cmd += " --status={status}".format(status=status)
    if comment is not None:
        cmd += " --comment='{comment}'".format(comment=comment)

    func = 'ligo.gracedb.rest.GraceDb.update_signoff'
    with mock.patch(func) as mock_cli_func:
        CLI(shlex.split(cmd))

    # Check call count
    assert mock_cli_func.call_count == 1

    # Get args used in function call
    cli_args, cli_kwargs = mock_cli_func.call_args

    # Check args used in function call
    assert cli_args == (cmd_args['s_id'], cmd_args['signoff_type'],)
    assert cli_kwargs == {'instrument': instrument or '', 'status': status,
                          'comment': comment}


SUPEREVENT_TEST_DATA = [
    (None, None, None, None),
    (1, 2, 3, 'G1234'),
]
@pytest.mark.parametrize("t_start,t_0,t_end,pref_ev",  # noqa: E302
                         SUPEREVENT_TEST_DATA)
def test_update_superevent_subcommand(CLI, t_start, t_0, t_end, pref_ev):
    """Test update superevent subcommand"""
    s_id = 'S001122a'

    # Generate command
    cmd = "update superevent {s_id}".format(s_id=s_id)
    if t_start is not None:
        cmd += " --t-start={ts}".format(ts=t_start)
    if t_0 is not None:
        cmd += " --t-0={t0}".format(t0=t_0)
    if t_end is not None:
        cmd += " --t-end={te}".format(te=t_end)
    if pref_ev is not None:
        cmd += " --preferred-event={pe}".format(pe=pref_ev)

    func = 'ligo.gracedb.rest.GraceDb.updateSuperevent'
    with mock.patch(func) as mock_cli_func:
        CLI(shlex.split(cmd))

    # Check call count
    assert mock_cli_func.call_count == 1

    # Get args used in function call
    cli_args, cli_kwargs = mock_cli_func.call_args

    # Check args used in function call
    assert cli_args == (s_id,)
    assert cli_kwargs == {'t_start': t_start, 't_0': t_0, 't_end': t_end,
                          'preferred_event': pref_ev}


GRBEVENT_TEST_DATA = [
    (None, None, None, None, None, None),
    (1, 2, 3, 4, 5, 'GRB000101A'),
    (1, None, 3, None, None, 'GRB000101123'),
]
@pytest.mark.parametrize(  # noqa: E302
    "ra,dec,error_radius,t90,redshift,designation",
    GRBEVENT_TEST_DATA
)
def test_update_grbevent_subcommand(
    CLI, ra, dec, error_radius, t90, redshift, designation
):
    """Test update grbevent subcommand"""
    gid = 'E123456'

    # Generate command
    cmd = "update grbevent {gid}".format(gid=gid)
    if ra is not None:
        cmd += " --ra={ra}".format(ra=ra)
    if dec is not None:
        cmd += " --dec={dec}".format(dec=dec)
    if error_radius is not None:
        cmd += " --error-radius={er}".format(er=error_radius)
    if t90 is not None:
        cmd += " --t90={t90}".format(t90=t90)
    if redshift is not None:
        cmd += " --redshift={rs}".format(rs=redshift)
    if designation is not None:
        cmd += " --designation={d}".format(d=designation)

    func = 'ligo.gracedb.rest.GraceDb.update_grbevent'
    with mock.patch(func) as mock_cli_func:
        CLI(shlex.split(cmd))

    # Check call count
    assert mock_cli_func.call_count == 1

    # Get args used in function call
    cli_args, cli_kwargs = mock_cli_func.call_args

    # Check args used in function call
    assert cli_args == (gid,)
    assert cli_kwargs == {'ra': ra, 'dec': dec, 'error_radius': error_radius,
                          't90': t90, 'redshift': redshift,
                          'designation': designation}



# Utils for command processing


def parse_delimited_string(data, delim=",", cast=None):
    """
    Parse a delimited string into a list. Can cast list elements to a type by
    setting 'cast' to a callable, like int.

    Examples:
        data='a,b,c,d' should give output ['a', 'b', 'c', 'd']
        data='1.2,1.4' should give output [1.2, 1.4] (with cast=float)
    """
    # Split data
    split_data = data.split(delim)

    # Cast data
    if cast is not None:
        split_data = [cast(element) for element in split_data]

    return split_data


def parse_delimited_string_or_single(data, delim=",", cast=None):
    """
    Parse a possibly-delimited string into a list or a single item.
    See docstring for 'parse_delimited_string' for more information.

    Examples:
        data='a,b,c,d' should give output ['a', 'b', 'c', 'd']
        data='one_item' should give output 'one_item'
        data='1.2,1.4' should give output [1.2, 1.4] (with cast=float)
        data='10.2' should give output 10.2 (with cast=float)
        data=None should give output None
    """
    if data is None:
        return None

    # Split and cast data
    split_data = parse_delimited_string(data, delim, cast)

    # Check if single element
    if (len(split_data) == 1):
        split_data = split_data[0]

    return split_data



# -*- coding: utf-8 -*-
# Copyright (C) Leo Singer, Brian Moe, Branson Stephens (2015)
#
# This file is part of gracedb
#
# gracedb is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# It is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with gracedb.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import absolute_import
from __future__ import print_function

"""
Some convenience logging classes courtesy of Leo Singer, provided as is.

Usage:
import logging
import ligo.gracedb.rest
import ligo.gracedb.logging

logging.basicConfig()
log = logging.getLogger('testing')

gracedb = ligo.gracedb.rest.GraceDb()
graceid = 'T62829'

log.addHandler(ligo.gracedb.logging.GraceDbLogHandler(gracedb, graceid))

# The following will create a log entry on the gracedb server
# (if the log level permits)
#
log.warn("this is a warning")
"""

import logging
import sys
import traceback
from six.moves.queue import Queue
from threading import Lock, Thread
__all__ = ('GraceDbLogHandler',)


_END_OF_STREAM = object()


class _GraceDbLogStream(object):

    def __init__(self, gracedb, graceid):
        self._closed = False
        self._lock = Lock()
        self._gracedb = gracedb
        self._graceid = graceid
        self._queue = Queue()
        self._thread = Thread(target=self._run)
        self._thread.daemon = True
        self._thread.start()

    def _run(self):
        while True:
            text = self._queue.get()
            if text is _END_OF_STREAM:
                self._queue.task_done()
                return
            try:
                self._gracedb.writeLog(self._graceid, text)
            except Exception:
                # If writeLog() fails, then print a message to stderr and keep
                # going. This is one rare circumstance where we must catch all
                # exceptions. If the logging thread crashes, then it will stop
                # clearing out the queue. Safe and rapid termination of the
                # thread requires that we ignore exceptions and continue
                # consuming items from the queue.
                if logging.raiseExceptions:
                    print('An error occurred in the log handler. '
                          'The traceback was:', file=sys.stderr)
                    traceback.print_exc()
            self._queue.task_done()

    def write(self, text):
        if text.strip():
            with self._lock:
                if self._closed:
                    raise RuntimeError('Log handler is already closed')
                self._queue.put(text)

    def close(self):
        with self._lock:
            if not self._closed:
                self._queue.put(_END_OF_STREAM)
                self._queue.join()
                self._thread.join()
                self._closed = True

    def __del__(self):
        self.close()


class _GraceDbLogFormatter(logging.Formatter):

    def __init__(self):
        logging.Formatter.__init__(self, logging.BASIC_FORMAT)

    def format(self, record):
        s = logging.Formatter.format(self, record)
        return '<div style="white-space:pre-wrap">' + s.strip("\n") + '</div>'


class GraceDbLogHandler(logging.StreamHandler):

    def __init__(self, gracedb, graceid):
        stream = _GraceDbLogStream(gracedb, graceid)
        logging.StreamHandler.__init__(self, stream)
        self.formatter = _GraceDbLogFormatter()

    def close(self):
        logging.StreamHandler.close(self)
        self.stream.close()

    def __del__(self):
        self.close()



import textwrap
import sys

from .base import RegisteredCommandBase, RegisteredSubCommandBase
from ..utils import parse_delimited_string


# Command registry - don't touch!
registry = []


###############################################################################
# Base command
###############################################################################
class SearchCommand(RegisteredCommandBase):
    name = "search"
    description = "Get a list of events or superevents based on a search query"
    subcommands = registry


###############################################################################
# Subcommands - registered to base command automatically
###############################################################################
class SearchChildBase(RegisteredSubCommandBase):
    _registry = registry


class SearchSupereventsCommand(SearchChildBase):
    name = "superevents"
    description = "Get a list of superevents based on a search query"
    long_description = textwrap.dedent("""\
        Get a list of superevents matching a given search query and
        print their attributes as a list of delimited columns.
    """).rstrip()
    default_columns = \
        "superevent_id,preferred_event,gw_events,labels,far,links.files"
    client_func = "superevents"

    def add_custom_arguments(self, parser):
        parser.add_argument(
            "query",
            type=str,
            help="Search query (surround with quotes)"
        )
        parser.add_argument(
            "--columns",
            type=parse_delimited_string,
            help=("Comma-separated list of parameters to show for each search "
                  "result. Use '.' to get nested parameters "
                  "(ex: 'links.files')"),
            default=self.default_columns
        )
        parser.add_argument(
            "--max-results",
            type=int,
            help="Maximum number of results to show",
            default=None
        )
        parser.add_argument(
            "--delimiter",
            type=str,
            default="TAB",
            help="Delimiter for output"
        )
        return parser

    def run(self, client, args):
        # Call client function to get iterator
        func = getattr(client, self.client_func)
        iterator = func(args.query, max_results=args.max_results)

        # Compile output
        output = []
        for item in iterator:
            output.append([self.process_element(item, col) for col in
                           args.columns])

        # Kludge for handling delimiter (if we set the default directly as
        # \t, it's not obvious in the help message)
        delim = args.delimiter
        if args.delimiter == 'TAB':
            delim = '\t'

        # Add title line
        output = [args.columns] + output

        # Join it all into a string
        out = '#' + "\n".join([delim.join(row) for row in output])
        return out

    def process_element(self, item, col):
        """
        Get values from nested dictionaries and join lists into
        comma-separated strings.
        """
        col_levels = col.split('.')
        a = item
        for c in col_levels:
            try:
                a = a[c]
            except KeyError:
                msg = ("'{col}' is not available in the response JSON. Check "
                       "the format on the server or by using '{prog} get' to "
                       "retrieve an individual event or superevent.") \
                    .format(col=col, prog=self.base_prog)
                print(msg)
                sys.exit(1)

        # Join lists into a comma-separated string
        if isinstance(a, list):
            a = ",".join(a)

        # Convert to string and return
        return str(a)


class SearchEventsCommand(SearchSupereventsCommand):
    name = "events"
    description = "Get a list of events based on a search query"
    long_description = textwrap.dedent("""\
        Get a list of events matching a given search query and
        print their attributes as a list of delimited columns.
    """).rstrip()
    default_columns = \
        "graceid,labels,group,pipeline,far,gpstime,created"
    client_func = "events"



try:
    from unittest import mock
except ImportError:  # python < 3
    import mock

import pytest


@pytest.mark.parametrize("voevent_number", [None, 12])
def test_superevent_voevents(safe_client, voevent_number):
    superevent_id = 'TS190302abc'

    # Set up templates mock
    mock_template = mock.MagicMock()
    if voevent_number:
        template_key = 'superevent-voevent-detail-template'
    else:
        template_key = 'superevent-voevent-list-template'
    mock_template_dict = {template_key: mock_template}
    template_prop = 'ligo.gracedb.rest.GraceDb.templates'
    with mock.patch('ligo.gracedb.rest.GraceDb.get') as mock_get, \
         mock.patch(template_prop, mock_template_dict):  # noqa: E127
        safe_client.voevents(superevent_id, voevent_num=voevent_number)

    get_call_args, get_call_kwargs = mock_get.call_args
    assert len(get_call_args) == 1
    assert get_call_kwargs == {}

    # Test template call kwargs
    num_kwargs = 1
    if voevent_number:
        num_kwargs += 1
    template_call_args, template_call_kwargs = mock_template.format.call_args
    assert template_call_args == ()
    assert len(template_call_kwargs) == num_kwargs
    assert template_call_kwargs['superevent_id'] == superevent_id
    if voevent_number:
        assert template_call_kwargs['N'] == voevent_number


@pytest.mark.parametrize("voevent_number", [None, 12])
def test_event_voevents(safe_client, voevent_number):
    graceid = 'T123456'

    # Set up templates mock
    mock_template = mock.MagicMock()
    if voevent_number:
        template_key = 'voevent-detail-template'
    else:
        template_key = 'voevent-list-template'
    mock_template_dict = {template_key: mock_template}
    template_prop = 'ligo.gracedb.rest.GraceDb.templates'
    with mock.patch('ligo.gracedb.rest.GraceDb.get') as mock_get, \
         mock.patch(template_prop, mock_template_dict):  # noqa: E127
        safe_client.voevents(graceid, voevent_num=voevent_number)

    get_call_args, get_call_kwargs = mock_get.call_args
    assert len(get_call_args) == 1
    assert get_call_kwargs == {}

    # Test template call kwargs
    num_kwargs = 1
    if voevent_number:
        num_kwargs += 1
    template_call_args, template_call_kwargs = mock_template.format.call_args
    assert template_call_args == ()
    assert len(template_call_kwargs) == num_kwargs
    assert template_call_kwargs['graceid'] == graceid
    if voevent_number:
        assert template_call_kwargs['N'] == voevent_number


@pytest.mark.parametrize("is_event", [True, False])
def test_create_voevent(safe_client, is_event):
    if is_event:
        obj_id = "T123456"
    else:
        obj_id = "TS121212abc"

    # Generate data
    voevent_type = 'PR'
    voevent_data = {
        'skymap_type': 'new',
        'skymap_filename': 'skymap.fits.gz',
        'internal': False,
        'open_alert': True,
        'hardware_inj': True,
        'CoincComment': False,
        'ProbHasNS': 0.6,
        'ProbHasRemnant': 0.7,
        'BNS': 0.1,
        'NSBH': 0.2,
        'BBH': 0.3,
        'Terrestrial': 0.4,
        'MassGap': 0.5,
    }

    # Set up templates mock
    mock_template = mock.MagicMock()
    if is_event:
        template_key = 'voevent-list-template'
    else:
        template_key = 'superevent-voevent-list-template'
    mock_template_dict = {template_key: mock_template}
    template_prop = 'ligo.gracedb.rest.GraceDb.templates'

    # Set up emgroup mock
    si_prop = 'ligo.gracedb.rest.GraceDb.service_info'
    mock_si_dict = {'voevent-types': {voevent_type: 'preliminary'}}

    with mock.patch('ligo.gracedb.rest.GraceDb.post') as mock_post, \
         mock.patch(si_prop, mock_si_dict), \
         mock.patch(template_prop, mock_template_dict):  # noqa: E127
        safe_client.createVOEvent(obj_id, voevent_type, **voevent_data)

    # Test call to self.post
    post_call_args, post_call_kwargs = mock_post.call_args
    assert len(post_call_args) == 1
    assert len(post_call_kwargs) == 1
    assert 'data' in post_call_kwargs
    request_body = post_call_kwargs['data']
    request_voevent_type = request_body.pop('voevent_type')
    assert request_voevent_type == voevent_type
    for k in voevent_data:
        assert request_body[k] == voevent_data[k]

    # Test template call kwargs
    template_call_args, template_call_kwargs = mock_template.format.call_args
    assert template_call_args == ()
    assert len(template_call_kwargs) == 1
    if is_event:
        call_key = 'graceid'
    else:
        call_key = 'superevent_id'
    assert template_call_kwargs[call_key] == obj_id


def test_create_voevent_bad_voevent_type(safe_client):
    # Set up templates mock
    mock_template = mock.MagicMock()
    mock_template_dict = {'superevent-voevent-list-template': mock_template}
    template_prop = 'ligo.gracedb.rest.GraceDb.templates'

    # Set up emgroup mock
    si_prop = 'ligo.gracedb.rest.GraceDb.service_info'
    mock_si_dict = {'voevent-types': {'PR': 'preliminary'}}

    err_msg = "voevent_type must be one of: {vts}".format(
        vts=",'".join(list(mock_si_dict['voevent-types'].values())))
    with mock.patch(si_prop, mock_si_dict), \
         mock.patch(template_prop, mock_template_dict):  # noqa: E127
        with pytest.raises(ValueError, match=err_msg):
            safe_client.createVOEvent('TS121212a', 'OTHER_TYPE')


def test_create_initial_voevent_no_skymap(safe_client):
    # Set up templates mock
    mock_template = mock.MagicMock()
    mock_template_dict = {'superevent-voevent-list-template': mock_template}
    template_prop = 'ligo.gracedb.rest.GraceDb.templates'

    # Set up emgroup mock
    si_prop = 'ligo.gracedb.rest.GraceDb.service_info'
    mock_si_dict = {'voevent-types': {'IN': 'initial'}}

    err_msg = "Skymap file is required for 'initial' VOEvents"
    with mock.patch(si_prop, mock_si_dict), \
         mock.patch(template_prop, mock_template_dict):  # noqa: E127
        with pytest.raises(ValueError, match=err_msg):
            safe_client.createVOEvent('TS121212a', 'IN')



# Tests for subcommands below 'list'
#  Ex: 'gracedb list labels'
import pytest
import shlex
try:
    from unittest import mock
except ImportError:
    import mock

# Apply module-level mark
pytestmark = pytest.mark.cli


###############################################################################
# Parametrized tests ##########################################################
###############################################################################
# We can parametrize most of the tests because these commands are simple
TEST_DATA = [
    ('emobservations', None),
    ('files', None),
    ('labels', None),
    ('logs', None),
    ('signoffs', None),
    ('voevents', None),
]
@pytest.mark.parametrize("name,cli_func", TEST_DATA)  # noqa: E302
def test_list_subcommands(CLI, name, cli_func):
    """Test most of the list subcommands"""
    s_id = 'S001122a'

    # Generate command
    cmd = 'list {name} {s_id}'.format(name=name, s_id=s_id)

    # Set up client function
    if cli_func is None:
        cli_func = name
    func = 'ligo.gracedb.rest.GraceDb.{func}'.format(func=cli_func)
    with mock.patch(func) as mock_cli_func:
        CLI(shlex.split(cmd))

    # Check call count
    assert mock_cli_func.call_count == 1

    # Get args used in function call
    cli_args, cli_kwargs = mock_cli_func.call_args

    # Check args used in function call
    assert cli_args == (s_id,)
    assert cli_kwargs == {}


###############################################################################
# Tests of individual subcommands #############################################
###############################################################################
def test_list_tags_subcommand(CLI):
    """Test list tags subcommand"""
    cmd_args = {
        's_id': 'S001122a',
        'N': 14,
    }

    # Generate command
    cmd = 'list tags {s_id} {N}'.format(**cmd_args)

    # Call command
    func = 'ligo.gracedb.rest.GraceDb.tags'
    with mock.patch(func) as mock_cli_func:
        CLI(shlex.split(cmd))

    # Check call count
    assert mock_cli_func.call_count == 1

    # Get args used in function call
    cli_args, cli_kwargs = mock_cli_func.call_args

    # Check args used in function call
    assert cli_args == (cmd_args['s_id'], cmd_args['N'],)
    assert cli_kwargs == {}



try:
    from unittest import mock
except ImportError:  # python < 3
    import mock
import six

import pytest


@pytest.mark.parametrize(
    "labels",
    [["BAD"], 'BAD', ["GOOD_LABEL", "BAD"], ('BAD', 'GOOD_LABEL')]
)
def test_creation_with_bad_labels(safe_client, labels):
    with mock.patch('ligo.gracedb.rest.GraceDb.allowed_labels',
                    new_callable=mock.PropertyMock) as mock_labels:
        mock_labels.return_value = ['GOOD_LABEL']

        with pytest.raises(ValueError):
            safe_client.createSuperevent(1, 2, 3, preferred_event='T0001',
                                         category='T', labels=labels)


@pytest.mark.parametrize("labels", [1, 2.34, {'test': 'bad'}, (1, 'GOOD')])
def test_creation_with_bad_label_type(safe_client, labels):

    with pytest.raises(TypeError):
        safe_client.createSuperevent(1, 2, 3, preferred_event='T0001',
                                     category='T', labels=labels)


@pytest.mark.parametrize("category", ['other', 'new', 'etc'])
def test_creation_with_bad_category(safe_client, category):
    with mock.patch('ligo.gracedb.rest.GraceDb.superevent_categories',
                    new_callable=mock.PropertyMock) as mock_sc:
        mock_sc.return_value = {
            'P': 'Production',
            'T': 'Test',
            'M': 'MDC',
        }

        with pytest.raises(ValueError):
            safe_client.createSuperevent(1, 2, 3, preferred_event='T0001',
                                         category=category)


@pytest.mark.parametrize("category", [1, 2.34, {'test': 'bad'}, None])
def test_creation_with_bad_category_type(safe_client, category):
    with mock.patch('ligo.gracedb.rest.GraceDb.superevent_categories',
                    new_callable=mock.PropertyMock) as mock_sc:
        mock_sc.return_value = {
            'P': 'Production',
            'T': 'Test',
            'M': 'MDC',
        }

        with pytest.raises(TypeError):
            safe_client.createSuperevent(1, 2, 3, preferred_event='T0001',
                                         category=category)
    pass


@pytest.mark.parametrize(
    "events",
    [1, ['G0001', 3], [()], [1, 2, 3.44]]
)
def test_creation_with_bad_event_graceids(safe_client, events):
    with pytest.raises(TypeError):
        safe_client.createSuperevent(1, 2, 3, preferred_event='T0001',
                                     category='T', events=events)


@pytest.mark.parametrize(
    "t_start,t_0,t_end,preferred_event,category,events,labels",
    [
        (1, 2, 3, 'T0001', 'T', None, None),
        (1, 2, 3, 'T0001', 'T', 'T0002', None),
        (1, 2, 3, 'T0001', 'T', None, 'INJ'),
        (1, 2, 3, 'T0001', 'T', ['T0002', 'T0003'], ['INJ', 'DQV']),
    ]
)
def test_creation_args(
    safe_client, t_start, t_0, t_end, preferred_event, category, events, labels
):
    with mock.patch('ligo.gracedb.rest.GraceDb.superevent_categories',
                    new_callable=mock.PropertyMock) as mock_sc, \
         mock.patch('ligo.gracedb.rest.GraceDb.allowed_labels',  # noqa: E127
                    new_callable=mock.PropertyMock) as mock_labels, \
         mock.patch('ligo.gracedb.rest.GraceDb.links',
                    new_callable=mock.PropertyMock), \
         mock.patch('ligo.gracedb.rest.GraceDb.post') as mock_post:
        mock_sc.return_value = {
            'P': 'Production',
            'T': 'Test',
            'M': 'MDC',
        }
        mock_labels.return_value = ['DQV', 'INJ']
        safe_client.createSuperevent(
            t_start, t_0, t_end, preferred_event, category=category,
            labels=labels, events=events
        )

    # Check args
    call_args, call_kwargs = mock_post.call_args
    body = call_kwargs['data']
    assert body['t_start'] == t_start
    assert body['t_0'] == t_0
    assert body['t_end'] == t_end
    assert body['preferred_event'] == preferred_event
    assert body['category'] == category
    if events:
        if isinstance(events, six.string_types):
            events = [events]
        assert body['events'] == events
    if labels:
        if isinstance(labels, six.string_types):
            labels = [labels]
        assert body['labels'] == labels


def test_update_no_args(safe_client):
    err_msg = ('Provide at least one of t_start, t_0, t_end, preferred_event, '
               'em_type, time_coinc_far, or space_coinc_far')
    with pytest.raises(ValueError, match=err_msg):
        safe_client.updateSuperevent('S181224a')


@pytest.mark.parametrize(
    "update_kwargs",
    [
        {'t_start': 1},
        {'t_0': 2},
        {'t_end': 3},
        {'preferred_event': 'G0001'},
        {'em_type': 'EM_TEST'},
        {'time_coinc_far': 4},
        {'space_coinc_far': 5},
        {'t_start': 111, 't_0': 222, 't_end': 333, 'preferred_event': 'G0001',
         'em_type': 'EM_TEST', 'time_coinc_far': 4, 'space_coinc_far': 5},
    ]
)
def test_update_args(safe_client, update_kwargs):
    with mock.patch('ligo.gracedb.rest.GraceDb.patch') as mock_patch, \
         mock.patch('ligo.gracedb.rest.GraceDb.templates'):  # noqa: E127
        safe_client.updateSuperevent('S181224a', **update_kwargs)

    # Check results
    call_args, call_kwargs = mock_patch.call_args
    body = call_kwargs['data']
    assert sorted(list(update_kwargs)) == sorted(list(body))
    for k in update_kwargs:
        assert update_kwargs[k] == body[k]


def test_add_event(safe_client):
    superevent_id = 'S190302abc'
    event_graceid = 'G123456'
    mock_template = mock.MagicMock()
    mock_template_dict = {'superevent-event-list-template': mock_template}
    template_prop = 'ligo.gracedb.rest.GraceDb.templates'
    with mock.patch('ligo.gracedb.rest.GraceDb.post') as mock_post, \
         mock.patch(template_prop, mock_template_dict):  # noqa: E127
        safe_client.addEventToSuperevent(superevent_id, event_graceid)

    call_args, call_kwargs = mock_post.call_args
    assert call_kwargs['data'] == {'event': event_graceid}

    template_call_args, template_call_kwargs = mock_template.format.call_args
    assert template_call_args == ()
    assert len(template_call_kwargs) == 1
    assert template_call_kwargs['superevent_id'] == superevent_id


def test_remove_event(safe_client):
    superevent_id = 'S190302abc'
    event_graceid = 'G123456'
    mock_template = mock.MagicMock()
    mock_template_dict = {'superevent-event-detail-template': mock_template}
    template_prop = 'ligo.gracedb.rest.GraceDb.templates'
    with mock.patch('ligo.gracedb.rest.GraceDb.delete') as mock_delete, \
         mock.patch(template_prop, mock_template_dict):  # noqa: E127
        safe_client.removeEventFromSuperevent(superevent_id, event_graceid)

    delete_call_args, delete_call_kwargs = mock_delete.call_args
    assert len(delete_call_args) == 1
    assert delete_call_kwargs == {}

    template_call_args, template_call_kwargs = mock_template.format.call_args
    assert template_call_args == ()
    assert len(template_call_kwargs) == 2
    assert template_call_kwargs['superevent_id'] == superevent_id
    assert template_call_kwargs['graceid'] == event_graceid


def test_confirm_as_gw(safe_client):
    superevent_id = 'S190302abc'
    mock_template = mock.MagicMock()
    mock_template_dict = {'superevent-confirm-as-gw-template': mock_template}
    template_prop = 'ligo.gracedb.rest.GraceDb.templates'
    with mock.patch('ligo.gracedb.rest.GraceDb.post') as mock_post, \
         mock.patch(template_prop, mock_template_dict):  # noqa: E127
        safe_client.confirm_superevent_as_gw(superevent_id)

    post_call_args, post_call_kwargs = mock_post.call_args
    assert len(post_call_args) == 1
    assert post_call_kwargs == {}

    template_call_args, template_call_kwargs = mock_template.format.call_args
    assert template_call_args == ()
    assert len(template_call_kwargs) == 1
    assert template_call_kwargs['superevent_id'] == superevent_id


# Test get
def test_get(safe_client):
    superevent_id = 'S190302abc'
    mock_template = mock.MagicMock()
    mock_template_dict = {'superevent-detail-template': mock_template}
    template_prop = 'ligo.gracedb.rest.GraceDb.templates'
    with mock.patch('ligo.gracedb.rest.GraceDb.get') as mock_get, \
         mock.patch(template_prop, mock_template_dict):  # noqa: E127
        safe_client.superevent(superevent_id)

    get_call_args, get_call_kwargs = mock_get.call_args
    assert len(get_call_args) == 1
    assert get_call_kwargs == {}

    template_call_args, template_call_kwargs = mock_template.format.call_args
    assert template_call_args == ()
    assert len(template_call_kwargs) == 1
    assert template_call_kwargs['superevent_id'] == superevent_id


# TODO: test search



try:
    from unittest import mock
except ImportError:  # python < 3
    import mock

import pytest


@pytest.mark.parametrize(
    "is_event,label_name",
    [
        (True, ""),
        (False, ""),
        (True, 'LABEL_NAME'),
        (False, 'LABEL_NAME'),
    ]
)
def test_get_labels(safe_client, is_event, label_name):
    if is_event:
        obj_id = 'T123456'
    else:
        obj_id = 'TS190302abc'

    # Set up templates mock
    mock_template = mock.MagicMock()
    if is_event:
        template_key = 'event-label-template'
    else:
        if label_name:
            template_key = 'superevent-label-detail-template'
        else:
            template_key = 'superevent-label-list-template'
    mock_template_dict = {template_key: mock_template}
    template_prop = 'ligo.gracedb.rest.GraceDb.templates'

    # Set up allowed_labels mock
    si_prop = 'ligo.gracedb.rest.GraceDb.service_info'
    mock_si_dict = {'labels': ['LABEL_NAME']}
    with mock.patch('ligo.gracedb.rest.GraceDb.get') as mock_get, \
         mock.patch(si_prop, mock_si_dict), \
         mock.patch(template_prop, mock_template_dict):  # noqa: E127
        safe_client.labels(obj_id, label=label_name)

    get_call_args, get_call_kwargs = mock_get.call_args
    assert len(get_call_args) == 1
    assert get_call_kwargs == {}

    # Test template call kwargs
    template_call_args, template_call_kwargs = mock_template.format.call_args
    assert template_call_args == ()
    if is_event:
        assert len(template_call_kwargs) == 2
        assert template_call_kwargs['graceid'] == obj_id
        assert template_call_kwargs['label'] == label_name
    else:
        if label_name:
            assert len(template_call_kwargs) == 2
            assert template_call_kwargs['superevent_id'] == obj_id
            assert template_call_kwargs['label_name'] == label_name
        else:
            assert len(template_call_kwargs) == 1
            assert template_call_kwargs['superevent_id'] == obj_id


def test_get_label_bad_label_name(safe_client):
    # Set up allowed_labels mock
    si_prop = 'ligo.gracedb.rest.GraceDb.service_info'
    mock_si_dict = {'labels': ['LABEL_NAME']}

    label = 'bad_label'
    err_msg = "Label '{label}' does not exist in the database".format(
        label=label)
    with mock.patch(si_prop, mock_si_dict):
        with pytest.raises(NameError, match=err_msg):
            safe_client.labels('T123456', label=label)


@pytest.mark.parametrize("is_event", [True, False])
def test_write_label(safe_client, is_event):
    label = 'TEST_LABEL'
    if is_event:
        obj_id = 'T123456'
        request_function = 'ligo.gracedb.client.GraceDBClient.put'
    else:
        obj_id = 'TS190302abc'
        request_function = 'ligo.gracedb.client.GraceDBClient.post'

    # Set up templates mock
    mock_template = mock.MagicMock()
    if is_event:
        template_key = 'event-label-template'
    else:
        template_key = 'superevent-label-list-template'
    mock_template_dict = {template_key: mock_template}
    template_prop = 'ligo.gracedb.rest.GraceDb.templates'

    # Set up allowed_labels mock
    si_prop = 'ligo.gracedb.rest.GraceDb.service_info'
    mock_si_dict = {'labels': [label]}
    with mock.patch(si_prop, mock_si_dict), \
         mock.patch(request_function) as mock_rf, \
         mock.patch(template_prop, mock_template_dict):  # noqa: E127
        safe_client.writeLabel(obj_id, label)

    rf_call_args, rf_call_kwargs = mock_rf.call_args
    assert len(rf_call_args) == 1
    assert len(rf_call_kwargs) == 1
    assert 'data' in rf_call_kwargs
    if is_event:
        assert rf_call_kwargs['data'] == {}
    else:
        assert rf_call_kwargs['data'] == {'name': label}

    # Test template call kwargs
    template_call_args, template_call_kwargs = mock_template.format.call_args
    assert template_call_args == ()
    if is_event:
        assert len(template_call_kwargs) == 2
        assert template_call_kwargs['graceid'] == obj_id
        assert template_call_kwargs['label'] == label
    else:
        assert len(template_call_kwargs) == 1
        assert template_call_kwargs['superevent_id'] == obj_id


def test_write_label_bad_label_name(safe_client):
    # Set up allowed_labels mock
    si_prop = 'ligo.gracedb.rest.GraceDb.service_info'
    mock_si_dict = {'labels': ['LABEL_NAME']}

    label = 'bad_label'
    err_msg = "Label '{label}' does not exist in the database".format(
        label=label)
    with mock.patch(si_prop, mock_si_dict):
        with pytest.raises(NameError, match=err_msg):
            safe_client.writeLabel('T123456', label)


@pytest.mark.parametrize("is_event", [True, False])
def test_remove_label(safe_client, is_event):
    label = 'TEST_LABEL'
    if is_event:
        obj_id = 'T123456'
    else:
        obj_id = 'TS190302abc'

    # Set up templates mock
    mock_template = mock.MagicMock()
    if is_event:
        template_key = 'event-label-template'
    else:
        template_key = 'superevent-label-detail-template'
    mock_template_dict = {template_key: mock_template}
    template_prop = 'ligo.gracedb.rest.GraceDb.templates'

    # Set up allowed_labels mock
    si_prop = 'ligo.gracedb.rest.GraceDb.service_info'
    mock_si_dict = {'labels': [label]}
    with mock.patch('ligo.gracedb.rest.GraceDb.delete') as mock_delete, \
         mock.patch(si_prop, mock_si_dict), \
         mock.patch(template_prop, mock_template_dict):  # noqa: E127
        safe_client.removeLabel(obj_id, label)

    delete_call_args, delete_call_kwargs = mock_delete.call_args
    assert len(delete_call_args) == 1
    assert len(delete_call_kwargs) == 0

    # Test template call kwargs
    template_call_args, template_call_kwargs = mock_template.format.call_args
    assert template_call_args == ()
    assert len(template_call_kwargs) == 2
    if is_event:
        assert template_call_kwargs['graceid'] == obj_id
        assert template_call_kwargs['label'] == label
    else:
        assert template_call_kwargs['superevent_id'] == obj_id
        assert template_call_kwargs['label_name'] == label



from __future__ import absolute_import, print_function
import json
import os
import six
import sys
import textwrap
from requests import Response

from ligo.gracedb.rest import GraceDb, DEFAULT_SERVICE_URL
from ligo.gracedb import __version__
from ligo.gracedb.exceptions import HTTPError
from ligo.gracedb.cli.commands.base import CommandBase
from ligo.gracedb.cli.commands.base import command_registry

# Version string
VERSION_STRING = "GraceDB Client {0}".format(__version__)


# Override and add a few methods to the base GraceDb client class
class CommandLineClient(GraceDb):
    # TP 2019: not sure if we still need to do this, but leaving it for now.
    # Hamstring 'adjustResponse' from the example REST client.
    # We don't want it messing with the response from the server.
    def adjustResponse(self, response):
        response.json = lambda: self.load_json_from_response(response)
        return response

    # TP 2019: not sure if we still need to override this from the GraceDb
    # class, but leaving it for now.
    # @classmethod
    # def output_and_die(cls, msg):
    #    sys.stderr.write(msg)
    #    sys.exit(1)


class CommandLineInterface(CommandBase):
    """Main class for base-level command-line interface"""
    subcommands = command_registry
    description = textwrap.dedent("""\
        Command-line interface to the ligo-gracedb client tools for
        interacting with the GraceDB API.
    """).rstrip()

    def __call__(self, args=None):
        if args is None:
            args = sys.argv[1:]

        # Parse args
        cmd, main_args, cmd_args = self.parse_args(args)

        # Set up client
        self.set_up_client(main_args)

        # Call subcommand and get output
        output = cmd(self.client, cmd_args)

        return output

    def add_custom_arguments(self, parser):
        # We set nargs='?' for handling legacy commands
        parser.add_argument("command", nargs='?')

        # Other options
        parser.add_argument(
            "-s",
            "--service-url",
            dest="service",
            type=str,
            metavar="URL",
            help="GraceDB service URL",
            default=os.environ.get("GRACEDB_SERVICE_URL", DEFAULT_SERVICE_URL)
        )
        parser.add_argument(
            "-p",
            "--proxy",
            dest="proxy",
            type=str,
            metavar="PROXY[:PORT]",
            help="HTTP Proxy",
            default=os.environ.get("HTTP_PROXY", None)
        )
        parser.add_argument(
            "-V",
            "--version",
            action='version',
            version=VERSION_STRING
        )
        parser.add_argument(
            "--output-type",
            dest='output_type',
            type=str,
            help=textwrap.dedent("""\
                Select output type: 'status' = status code
                only, 'json' = full response JSON. Doesn't
                apply to some some commands which have
                pre-defined outputs, like 'info', 'ping',
                'search', etc.""").rstrip(),
            default='json', choices=['status', 'json']
        )
        parser.add_argument(
            '--username',
            dest='username',
            type=str,
            help='Basic auth username',
            default=None
        )
        parser.add_argument(
            '--password',
            dest='password',
            type=str,
            help='Basic auth password',
            default=None
        )
        parser.add_argument(
            '--creds',
            dest='creds',
            type=str,
            help=textwrap.dedent("""\
                Paths to certficate file and key file
                (comma-separated) OR path to single combined
                proxy file. Used for X.509 authentication
            """).rstrip(),
            default=None
        )
        parser.add_argument(
            "-n",
            "--force-noauth",
            dest="force_noauth",
            action="store_true",
            help="Do not use any authentication credentials",
            default=False
        )
        parser.add_argument(
            "-f",
            "--fail-if-noauth",
            dest="fail_if_noauth",
            action="store_true",
            help="Fail if no authentication credentials are found",
            default=False
        )
        return parser

    # Utils -------------------------------------------------------------------
    def set_up_client(self, args):
        # Handle proxy args
        proxy = args.proxy
        proxyport = None
        if proxy and proxy.find(':') > 0:
            try:
                proxy, proxyport = proxy.split(':')
                proxyport = int(proxyport)
            except Exception:
                print("Malformed proxy: '{0}'".format(proxy))
                sys.exit(1)

        # Handle creds arg
        creds = args.creds
        if creds:
            creds = creds.split(",")
            if len(creds) == 1:
                # Combined proxy file
                creds = creds[0]
            elif len(creds) == 2:
                # Cert file and key file
                pass
            else:
                print("Malformed 'creds' argument: {0}".format(args.creds))
                sys.exit(1)

        # Define kwargs for initializing client
        client_kwargs = {
            'service_url': args.service,
            'force_noauth': args.force_noauth,
            'username': args.username,
            'password': args.password,
            'fail_if_noauth': args.fail_if_noauth,
            'cred': creds,
        }

        # Initialize client
        self.client = CommandLineClient(**client_kwargs)


# Define a function for the entry_point
def main(args=None):
    if args is None:
        args = sys.argv[1:]

    cli = CommandLineInterface()

    # Try to test response and handle errors
    try:
        response = cli(args)
    except HTTPError as e:
        print('Error: {code} {reason}. {text}.'.format(
            code=e.status_code,
            reason=e.reason,
            text=e.text))
        sys.exit(1)
    except Exception as e:
        print('Error: {msg}'.format(msg=str(e)))
        sys.exit(1)

    if isinstance(response, Response):
        if (cli.args.output_type == 'json'):
            # Handle errors
            if response.status_code >= 400:
                output = '{code} {reason}'.format(code=response.status_code,
                                                  reason=response.reason)
                # Only add message if it's not really long (i.e., it's not
                # an HTML error page)
                msg = response.text
                if isinstance(msg, bytes):
                    msg = msg.decode()
                if (len(msg) < 1000):
                    output += ': {msg}'.format(msg=msg)
                print(output)
                sys.exit(1)
            # Handle errors raised in load_json_or_die()
            if response.status_code == 204:
                output = {}
                print(output)
                sys.exit(1)
            else:
                try:
                    output = response.json()
                except Exception as e:
                    print(str(e))
                    sys.exit(1)

            print(json.dumps(output, indent=4))
        elif (cli.args.output_type == 'status'):
            print('Server returned {status}: {reason}'.format(
                status=response.status_code, reason=response.reason))
    elif isinstance(response, dict):
        print(json.dumps(response, indent=4))
    elif isinstance(response, six.string_types):
        print(response)
    elif isinstance(response, bytes):
        print(response.decode())
    else:
        print("Unexpected response type {tp}".format(tp=type(response)))
        print("Response: {resp}".format(resp=str(response)))
        sys.exit(1)


if __name__ == "__main__":
    main()



import pytest

from ligo.gracedb.cli.client import CommandLineInterface, main


@pytest.fixture
def expected_subcommands():
    """Names of expected subcommands for CLI"""
    return ['add', 'create', 'delete', 'get', 'list', 'remove', 'search',
            'update', 'confirm_as_gw', 'expose', 'hide', 'ping', 'info',
            'credentials']


@pytest.fixture
def CLI():
    """Instantiated command-line interface"""
    return CommandLineInterface()


@pytest.fixture
def main_tester():
    """Get main() function which is used as a wrapper for the CLI"""
    return main



import os
import stat

import pytest

from .. import netrc, NetrcParseError

try:
    FileNotFoundError
except NameError:
    FileNotFoundError = IOError


def check(filename):
    # Check that an exception is raised when the netrc file does not exist.
    with pytest.raises(FileNotFoundError) as excinfo:
        netrc()
    assert excinfo.value.filename == filename

    with open(filename, 'w') as f:
        fd = f.fileno()
        # Check that an exception is raised when the permissions too relaxed.
        os.fchmod(fd, stat.S_IRWXU | stat.S_IRWXG | stat.S_IRWXO)
        with pytest.raises(NetrcParseError) as excinfo:
            netrc()
        assert excinfo.value.filename == filename

        # Check that no exceptions are raised if the permissions are correct.
        os.fchmod(fd, stat.S_IRWXU)
        netrc()


def test_netrc_in_environ(monkeypatch, tmpdir):
    """Check file specified by NETRC environment variable"""
    filename = str(tmpdir / 'netrc')
    monkeypatch.setenv('NETRC', filename)
    check(filename)


def test_netrc_in_homedir(monkeypatch, tmpdir):
    """Check file specified by home directory"""
    filename = str(tmpdir / '.netrc')
    monkeypatch.setenv('HOME', str(tmpdir))
    check(filename)



import os
import shutil
import textwrap

from .base import RegisteredCommandBase, RegisteredSubCommandBase
from ..parsers import object_id_parser, graceid_parser, superevent_id_parser


# Command registry - don't touch!
registry = []


###############################################################################
# Base command
###############################################################################
class GetCommand(RegisteredCommandBase):
    name = "get"
    description = textwrap.dedent("""\
        Download a file or get information about a log entry, label,
        EM observation, VOEvent, signoff, event, or superevent
    """).rstrip()
    subcommands = registry


###############################################################################
# Subcommands - registered to base command automatically
###############################################################################
class GetChildBase(RegisteredSubCommandBase):
    _registry = registry


class GetEmobservationCommand(GetChildBase):
    name = "emobservation"
    description = textwrap.dedent("""\
        Retrieve information about a single EM observation associated with an
        event or superevent
    """).rstrip()
    parent_parsers = (object_id_parser,)

    def add_custom_arguments(self, parser):
        parser.add_argument(
            "N",
            type=int,
            help="Index number of the EM observation"
        )
        return parser

    def run(self, client, args):
        return client.emobservations(args.object_id,
                                     emobservation_num=args.N)


class GetEventCommand(GetChildBase):
    name = "event"
    description = "Retrieve information about a single event"
    parent_parsers = (graceid_parser,)

    def run(self, client, args):
        return client.event(args.graceid)


class GetFileCommand(GetChildBase):
    name = "file"
    description = "Download a file associated with an event or superevent"
    parent_parsers = (object_id_parser,)

    def add_custom_arguments(self, parser):
        parser.add_argument(
            "filename",
            type=str,
            help="Name of file (on the server) to download"
        )
        parser.add_argument(
            "destination",
            type=str,
            nargs='?',
            help=("Path to save file at. Default is '-', which prints "
                  "the contents to stdout"),
            default='-'
        )
        return parser

    def run(self, client, args):
        # Check that we *could* write the file before we go to the trouble
        # of getting it.  Also, try not to open a file until we know that we
        # we have data.
        if args.destination != '-':
            full_dir_path = os.path.dirname(os.path.abspath(args.destination))
            if not os.access(full_dir_path, os.W_OK):
                raise IOError("{0}: permission denied".format(
                              args.destination))

        # Get file content
        response = client.files(args.object_id, args.filename)

        # Handle response
        if response.status_code == 200:
            if args.destination == '-':
                # For stdout, return string contents of file
                file_contents = response.read()
                return file_contents
            else:
                # Otherwise, save to file and return string message
                with open(args.destination, 'wb') as fh:
                    shutil.copyfileobj(response, fh)
                return "File '{fname}' for {obj_id} saved at {path}".format(
                    fname=args.filename, obj_id=args.object_id,
                    path=args.destination)
        else:
            # On error, return full response
            return response


class GetLabelCommand(GetChildBase):
    name = "label"
    description = \
        "Get information about a label attached to an event or superevent"
    parent_parsers = (object_id_parser,)

    def add_custom_arguments(self, parser):
        parser.add_argument("label", type=str, help="Label name")
        return parser

    def run(self, client, args):
        return client.labels(args.object_id, label=args.label)


class GetLogCommand(GetChildBase):
    name = "log"
    description = "Retrieve an event or superevent log entry"
    parent_parsers = (object_id_parser,)

    def add_custom_arguments(self, parser):
        parser.add_argument(
            "N",
            type=int,
            help="Index number of the log entry"
        )
        return parser

    def run(self, client, args):
        return client.logs(args.object_id, log_number=args.N)


class GetSignoffCommand(GetChildBase):
    name = "signoff"
    description = "Retrieve a superevent signoff"
    long_description = textwrap.dedent("""\
        Retrieve a signoff attached to a superevent. Event signoff retrieval
        is not presently implemented.
    """).rstrip()
    parent_parsers = (superevent_id_parser,)

    def add_custom_arguments(self, parser):
        parser.add_argument(
            "signoff_type",
            type=str,
            help=("Signoff type (do '{prog} info signoff_types' to see "
                  "options)").format(prog=self.base_prog)
        )
        parser.add_argument(
            "instrument",
            type=str,
            nargs='?',
            help=("Instrument code (do '{prog} info instruments' to see "
                  "options). Required for operator signoffs.")
            .format(prog=self.base_prog)
        )
        return parser

    def run(self, client, args):
        instrument = args.instrument or ''  # Convert None to ''
        return client.signoffs(
            args.superevent_id, signoff_type=args.signoff_type,
            instrument=instrument
        )


class GetSupereventCommand(GetChildBase):
    name = "superevent"
    description = "Retrieve information about a single superevent"
    parent_parsers = (superevent_id_parser,)

    def run(self, client, args):
        return client.superevent(args.superevent_id)


class GetVoeventCommand(GetChildBase):
    name = "voevent"
    description = "Retrieve a VOEvent associated with an event or superevent"
    parent_parsers = (object_id_parser,)

    def add_custom_arguments(self, parser):
        parser.add_argument("N", type=int, help="Index number of the VOEvent")
        return parser

    def run(self, client, args):
        return client.voevents(args.object_id, voevent_num=args.N)



# Tests for subcommands below 'remove'
#  Ex: 'gracedb remove event'
import pytest
import shlex
try:
    from unittest import mock
except ImportError:
    import mock

# Apply module-level mark
pytestmark = pytest.mark.cli


###############################################################################
# Tests of individual subcommands #############################################
###############################################################################
def test_remove_event_subcommand(CLI):
    """Test remove event subcommand"""
    s_id = 'S001122a'
    g_id = 'G123456'
    cmd = 'remove event {s_id} {g_id}'.format(s_id=s_id, g_id=g_id)

    func = 'ligo.gracedb.rest.GraceDb.removeEventFromSuperevent'
    with mock.patch(func) as mock_cli_func:
        CLI(shlex.split(cmd))

    # Check call count
    assert mock_cli_func.call_count == 1

    # Get args used in function call
    cli_args, cli_kwargs = mock_cli_func.call_args

    # Check args used in function call
    assert cli_args == (s_id, g_id,)
    assert cli_kwargs == {}


def test_remove_label_subcommand(CLI):
    """Test remove label subcommand"""
    s_id = 'S001122a'
    label_name = 'TEST_LABEL'
    cmd = 'remove label {s_id} {label}'.format(s_id=s_id, label=label_name)

    func = 'ligo.gracedb.rest.GraceDb.removeLabel'
    with mock.patch(func) as mock_cli_func:
        CLI(shlex.split(cmd))

    # Check call count
    assert mock_cli_func.call_count == 1

    # Get args used in function call
    cli_args, cli_kwargs = mock_cli_func.call_args

    # Check args used in function call
    assert cli_args == (s_id, label_name,)
    assert cli_kwargs == {}


def test_remove_tag_subcommand(CLI):
    """Test remove tag subcommand"""
    s_id = 'S001122a'
    log_N = 12
    tag_name = 'new_tag'
    cmd = 'remove tag {s_id} {N} {tag}'.format(
        s_id=s_id, N=log_N,
        tag=tag_name
    )

    func = 'ligo.gracedb.rest.GraceDb.removeTag'
    with mock.patch(func) as mock_cli_func:
        CLI(shlex.split(cmd))

    # Check call count
    assert mock_cli_func.call_count == 1

    # Get args used in function call
    cli_args, cli_kwargs = mock_cli_func.call_args

    # Check args used in function call
    assert cli_args == (s_id, log_N, tag_name,)
    assert cli_kwargs == {}



import pytest
try:
    from unittest import mock
except ImportError:  # py < 3
    import mock


def test_update_grbevent_no_args_provided(safe_client):
    err_str = ('Provide at least one of ra, dec, error_radius, t90, '
               'redshift, or designation')
    with pytest.raises(ValueError, match=err_str):
        safe_client.update_grbevent('E123456')


GRBEVENT_DATA = [
    (1, None, None, None, None, None),
    (None, 2, 3, None, 5, 'test'),
    (1, 2, 3, 4, 5, 'test'),
]
@pytest.mark.parametrize(  # noqa: E302
    "ra,dec,error_radius,t90,redshift,designation",
    GRBEVENT_DATA
)
def test_update_grbevent_with_args(
    safe_client, ra, dec, error_radius, t90, redshift, designation
):

    with mock.patch.object(safe_client, 'patch') as mock_patch, \
         mock.patch('ligo.gracedb.rest.GraceDb.templates',  # noqa: E127
                    new_callable=mock.PropertyMock):
        safe_client.update_grbevent(
            'E123456', ra=ra, dec=dec, error_radius=error_radius, t90=t90,
            redshift=redshift, designation=designation
        )

    # Check calls to patch() method
    call_args, call_kwargs = mock_patch.call_args
    assert len(call_kwargs) == 1
    assert 'data' in call_kwargs

    # Construct expected body content
    expected_body = {}
    if ra is not None:
        expected_body['ra'] = ra
    if dec is not None:
        expected_body['dec'] = dec
    if error_radius is not None:
        expected_body['error_radius'] = error_radius
    if t90 is not None:
        expected_body['t90'] = t90
    if redshift is not None:
        expected_body['redshift'] = redshift
    if designation is not None:
        expected_body['designation'] = designation

    assert call_kwargs['data'] == expected_body



# Tests for intermediate commands
#  Ex: 'gracedb confirm_as_gw' because it has no subcommands.
#  NOT 'gracedb add' because it has subcommands like 'gracedb add event'
import pytest
import six
try:
    from unittest import mock
except ImportError:
    import mock

from ligo.gracedb.cli.commands.subcommands import InfoCommand

# Apply module-level mark
pytestmark = pytest.mark.cli


###############################################################################
# Test data ###################################################################
###############################################################################
SUBCOMMANDS_TEST_DATA = [
    "confirm_as_gw", "credentials", "expose", "hide", "info", "ping"
]


###############################################################################
# Parametrized tests ##########################################################
###############################################################################
@pytest.mark.parametrize("cmd", SUBCOMMANDS_TEST_DATA)
def test_subcommand_help(CLI, cmd):
    """Test level 1 subcommand help functionality"""
    command_line_cmd = '{cmd} --help'.format(cmd=cmd)
    print_help_func_str = \
        'ligo.gracedb.cli.parsers.argparse.ArgumentParser.print_help'
    with mock.patch(print_help_func_str) as mock_print_help:
        with pytest.raises(SystemExit) as excinfo:
            CLI(command_line_cmd.split(' '))

    # Make sure print_help is called once
    assert mock_print_help.call_count == 1

    # Check exit code
    assert excinfo.value.code == 0


@pytest.mark.parametrize("cmd", SUBCOMMANDS_TEST_DATA)
def test_subcommand_help_content(CLI, cmd):
    """Check level 1 subcommand help content"""

    # Create an instance of the subcommand
    sc_instance = CLI.subcommand_dict[cmd](parent_prog=CLI.parser.prog)

    # Get help dialogue
    help_text = sc_instance.parser.format_help()

    # Check usage
    assert help_text.startswith('usage: {prog} {sc}'.format(
        prog=CLI.parser.prog, sc=cmd))


###############################################################################
# Tests of individual subcommands #############################################
###############################################################################
def test_confirm_as_gw_subcommand(CLI):
    """Test confirm_as_gw subcommand"""
    s_id = 'S001122a'

    func = 'ligo.gracedb.rest.GraceDb.confirm_superevent_as_gw'
    with mock.patch(func) as mock_cli_func:
        CLI(['confirm_as_gw', s_id])

    # Check call count
    assert mock_cli_func.call_count == 1

    # Get args used in function call
    cli_args, cli_kwargs = mock_cli_func.call_args

    # Check args used in function call
    assert cli_args == (s_id,)
    assert cli_kwargs == {}


CREDENTIALS_TEST_DATA = [
    ('client', 'show_credentials', {'print_output': False}),
    ('server', 'get_user_info', {}),
]
@pytest.mark.parametrize("cmd,cli_func,kw",  # noqa: E302
                         CREDENTIALS_TEST_DATA)
def test_credentials_subcommand(CLI, cmd, cli_func, kw):
    """Test credentials subcommand"""
    func = 'ligo.gracedb.rest.GraceDb.{0}'.format(cli_func)
    with mock.patch(func) as mock_cli_func:
        CLI(['credentials', cmd])

    # Check call count
    assert mock_cli_func.call_count == 1

    # Get args used in function call
    cli_args, cli_kwargs = mock_cli_func.call_args

    # Check args used in function call
    assert cli_args == ()
    assert cli_kwargs == kw


def test_expose_subcommand(CLI):
    """Test expose subcommand"""
    s_id = 'S001122a'

    func = 'ligo.gracedb.rest.GraceDb.modify_permissions'
    with mock.patch(func) as mock_cli_func:
        CLI(['expose', s_id])

    # Check call count
    assert mock_cli_func.call_count == 1

    # Get args used in function call
    cli_args, cli_kwargs = mock_cli_func.call_args

    # Check args used in function call
    assert cli_args == (s_id, 'expose')
    assert cli_kwargs == {}


def test_hide_subcommand(CLI):
    """Test hide subcommand"""
    s_id = 'S001122a'

    func = 'ligo.gracedb.rest.GraceDb.modify_permissions'
    with mock.patch(func) as mock_cli_func:
        CLI(['hide', s_id])

    # Check call count
    assert mock_cli_func.call_count == 1

    # Get args used in function call
    cli_args, cli_kwargs = mock_cli_func.call_args

    # Check args used in function call
    assert cli_args == (s_id, 'hide')
    assert cli_kwargs == {}


def test_ping_subcommand(CLI):
    """Test ping subcommand"""
    func = 'ligo.gracedb.rest.GraceDb.ping'
    with mock.patch(func) as mock_cli_func:
        mock_cli_func.return_value = mock.MagicMock(status_code=123)
        output = CLI(['ping'])

    # Check call count
    assert mock_cli_func.call_count == 1

    # Get args used in function call
    cli_args, cli_kwargs = mock_cli_func.call_args

    # Check args used in function call
    assert cli_args == ()
    assert cli_kwargs == {}

    # Check output
    assert output == 'Response from {server}: {status}'.format(
        server=CLI.client._service_url, status=mock_cli_func().status_code)


@pytest.mark.parametrize(
    "cmd,cli_func",
    list(six.iteritems(InfoCommand.options))
)
def test_info_subcommand(CLI, cmd, cli_func):
    """Test info subcommand"""

    func = 'ligo.gracedb.rest.GraceDb.{func}'.format(func=cli_func)
    with mock.patch(func, new_callable=mock.PropertyMock) as mock_cli_func:
        mock_cli_func.return_value = ['a', 'b', 'c', 'd']
        output = CLI(['info', cmd])

    # Check call count
    assert mock_cli_func.call_count == 1

    # Get args used in function call
    cli_args, cli_kwargs = mock_cli_func.call_args

    # Check args used in function call
    assert cli_args == ()
    assert cli_kwargs == {}

    # Check output
    assert output == ", ".join(mock_cli_func())



# -*- coding: utf-8 -*-
# Copyright (C) Brian Moe, Branson Stephens (2015)
#
# This file is part of gracedb
#
# gracedb is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# It is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with gracedb.  If not, see <http://www.gnu.org/licenses/>.

import os
import re
import sys
from setuptools import setup, find_packages


def parse_version(path):
    """Extract the `__version__` string from the given file"""
    with open(path, 'r') as fp:
        version_file = fp.read()
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]",
                              version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")


# Get readme file for long_description
readme_file = os.path.join(os.path.dirname(__file__), 'README.rst')
with open(readme_file, 'rb') as f:
    long_description = f.read().decode().strip()


# Required packages for tests
tests_require = []
# Add pytest requirement - 5.0.0+ only supports Python 3.5+
pytest_requirement = 'pytest>=3.1.0'
pyparsing_requirement = 'pyparsing==2.4.7'
zipp_requirement = 'zipp==0.6.0'
it_requirement = 'more-itertools==5.0.0'
if sys.version_info < (3, 5):
    pytest_requirement += ',<5.0.0'
tests_require.append(pytest_requirement)
tests_require.append(pyparsing_requirement)
tests_require.append(zipp_requirement)
tests_require.append(it_requirement)
tests_require.append('pytest-cov')

# Add mock for Python 2
if sys.version_info.major < 3:
    tests_require.append('mock==3.0.5')

# Only install setup_requires for the specific command being used
SETUP_REQUIRES = []
if 'test' in sys.argv:
    SETUP_REQUIRES.append('pytest-runner>=2.12')

# Classifiers
CLASSIFIERS = [
    'Development Status :: 4 - Beta',
    'Intended Audience :: Developers',
    'Intended Audience :: End Users/Desktop',
    'Intended Audience :: Science/Research',
    ('License :: OSI Approved :: GNU General Public License v3 or later '
        '(GPLv3+)'),
    'Operating System :: POSIX',
    'Programming Language :: Python :: 2',
    'Programming Language :: Python :: 2.7',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3.5',
    'Programming Language :: Python :: 3.6',
    'Programming Language :: Python :: 3.7',
    'Topic :: Scientific/Engineering',
    'Topic :: Scientific/Engineering :: Astronomy',
    'Topic :: Scientific/Engineering :: Physics',
]


###############################################################################
# Call setup() ################################################################
###############################################################################
setup(
    name="ligo-gracedb",
    version=parse_version(os.path.join('ligo', 'gracedb', 'version.py')),
    author=("Tanner Prestegard, Alexander Pace, Branson Stephens, Brian Moe, "
            "Patrick Brady"),
    author_email="tanner.prestegard@ligo.org, alexander.pace@ligo.org",
    description="A Python package for accessing the GraceDB API.",
    long_description=long_description,
    long_description_content_type='text/x-rst',
    url="https://git.ligo.org/lscsoft/gracedb-client",
    license='GPL-3.0-or-later',
    namespace_packages=['ligo'],
    packages=find_packages(),
    classifiers=CLASSIFIERS,
    install_requires=['future>=0.15.0',
                      'six>=1.9.0',
                      'cryptography>=1.7.2',
                      'requests>=2.6.0',
                      ],
    setup_requires=SETUP_REQUIRES,
    tests_require=tests_require,
    package_data={
        'ligo.gracedb.test': [
            'integration/data/*',
        ],
    },
    entry_points={
        'console_scripts': [
            'gracedb=ligo.gracedb.cli.client:main',
            'gracedb_legacy=ligo.gracedb.legacy_cli:main',
        ],
    },
)



import six
import textwrap

from .base import RegisteredSubCommandBase
from ..parsers import superevent_id_parser


class ConfirmAsGwCommand(RegisteredSubCommandBase):
    name = "confirm_as_gw"
    description = "Confirm a superevent as a GW"
    long_description = textwrap.dedent("""\
        Confirm a superevent as a gravitational wave. Specific permissions are
        required to perform this action on non-Test superevents.
    """).rstrip()
    parent_parsers = (superevent_id_parser,)

    def run(self, client, args):
        return client.confirm_superevent_as_gw(args.superevent_id)


class CredentialsCommand(RegisteredSubCommandBase):
    name = "credentials"
    description = "Display your credentials"
    long_description = textwrap.dedent("""\
        Display the credentials that the client will use to make API requests
        or get information about your user account on the server. Useful for
        debugging and ensuring that your credentials and user account
        information are what you expect them to be.
    """).rstrip()

    def add_custom_arguments(self, parser):
        parser.add_argument(
            'type',
            type=str,
            choices=['client', 'server'],
            help=('View credentials found by the client or get your user '
                  'account information from the server')
        )
        return parser

    def run(self, client, args):
        if (args.type == 'client'):
            return client.show_credentials(print_output=False)
        else:
            return client.get_user_info()


class ExposeCommand(RegisteredSubCommandBase):
    name = "expose"
    description = "Expose a superevent to non-internal users"
    long_description = textwrap.dedent("""\
        Expose a superevent to LV-EM and public users. Special permissions
        are required to perform this action.
    """).rstrip()

    def add_custom_arguments(self, parser):
        parser.add_argument(
            'superevent_id',
            type=str,
            help="ID of the superevent to expose"
        )
        return parser

    def run(self, client, args):
        return client.modify_permissions(args.superevent_id, 'expose')


class HideCommand(RegisteredSubCommandBase):
    name = "hide"
    description = \
        "Make an exposed superevent accessible to internal users only"
    long_description = textwrap.dedent("""\
        Make a superevent accessible to internal users only. Specific
        permissions are required to perform this action.
    """).rstrip()

    def add_custom_arguments(self, parser):
        parser.add_argument(
            'superevent_id',
            type=str,
            help="ID of the superevent to hide"
        )
        return parser

    def run(self, client, args):
        return client.modify_permissions(args.superevent_id, 'hide')


class PingCommand(RegisteredSubCommandBase):
    name = "ping"
    description = \
        "Make a basic request to check your connection to the server"

    def run(self, client, args):
        response = client.ping()
        output = 'Response from {server}: {status}'.format(
            server=client._service_url, status=response.status_code)
        if (response.status_code == 200):
            output += ' OK'
        return output


class InfoCommand(RegisteredSubCommandBase):
    """
    These commands all just print basic information obtained from the API root
    and attached to the client instance as a property. Each dict in 'options'
    corresponds to a command.
    """
    name = "info"
    description = "Get information from the server"
    long_description = textwrap.dedent("""\
        Get available EM groups, LVC analysis groups, pipelines, searches,
        labels, signoff types, signoff statuses, superevent categories,
        VOEvent types, or server code version from the server
    """).rstrip()

    # Mapping from argument value to client property
    options = {
        'emgroups': 'em_groups',
        'groups': 'groups',
        'instruments': 'instruments',
        'labels': 'allowed_labels',
        'pipelines': 'pipelines',
        'searches': 'searches',
        'server_version': 'server_version',
        'signoff_statuses': 'signoff_statuses',
        'signoff_types': 'signoff_types',
        'superevent_categories': 'superevent_categories',
        'voevent_types': 'voevent_types',
    }

    def add_custom_arguments(self, parser):
        parser.add_argument(
            "items",
            type=str,
            choices=sorted(list(self.options)),
            help="Information to display"
        )
        return parser

    def run(self, client, args):
        # Get list of objects from server, sort, and print
        data = getattr(client, self.options.get(args.items))

        # Convert dict to list of key (value) strings
        if isinstance(data, dict):
            data = ['{k} ({v})'.format(k=k, v=v) for k, v in
                    six.iteritems(data)]

        # Join list into comma-separated string
        if isinstance(data, list):
            data = ", ".join(sorted(data))

        # Handle None response
        if data is None:
            data = 'Data not found on server.'

        return data


class ShowCommand(InfoCommand):
    """Legacy version of InfoCommand. Same functionality"""
    legacy = True
    name = "show"
    description = "DEPRECATED: see 'gracedb info'"
    long_description = description



import os

import pytest

from ligo.gracedb.rest import GraceDb


@pytest.fixture
def client():
    """A full client instance for use in integration tests"""
    service_url = os.environ.get(
        'TEST_SERVICE',
        'https://gracedb-test.ligo.org/api/'
    )
    return GraceDb(service_url=service_url)


@pytest.fixture
def test_data_dir():
    d = os.environ.get(
        'TEST_DATA_DIR',
        os.path.join(os.path.dirname(__file__), 'data')
    )
    return d


@pytest.fixture
def create_event(client, test_data_dir):
    def _inner(pipeline='gstlal', search='LowMass',
               filename=os.path.join(test_data_dir, 'cbc-lm.xml')):
        response = client.createEvent('Test', pipeline, filename,
                                      search=search)
        return response.json()
    return _inner


@pytest.fixture
def create_superevent(client, test_data_dir, create_event):
    def _inner(
        pipeline='gstlal', search='LowMass',
        filename=os.path.join(test_data_dir, 'cbc-lm.xml'),
        t_start=1, t_0=2, t_end=3
    ):
        event = create_event(pipeline=pipeline, search=search,
                             filename=filename)
        response = client.createSuperevent(t_start, t_0, t_end,
                                           event['graceid'], category='T')
        return response.json()
    return _inner


@pytest.fixture
def create_obj(client, test_data_dir, create_event, create_superevent):
    def _inner(
        obj_type='event', pipeline='gstlal', search='LowMass',
        filename=os.path.join(test_data_dir, 'cbc-lm.xml'),
        t_start=1, t_0=2, t_end=3
    ):
        event = create_event(pipeline=pipeline, search=search,
                             filename=filename)
        if obj_type == 'superevent':
            response = client.createSuperevent(t_start, t_0, t_end,
                                               event['graceid'], category='T')
            obj = response.json()
            obj_id = obj['superevent_id']
        elif obj_type == 'event':
            obj = event
            obj_id = event['graceid']
        else:
            raise ValueError("obj_type must be 'event' or 'superevent'")
        return obj, obj_id
    return _inner



# Tests for subcommands below 'delete'
#  Ex: 'gracedb delete signoff'
import pytest
import shlex
try:
    from unittest import mock
except ImportError:
    import mock

# Apply module-level mark
pytestmark = pytest.mark.cli


###############################################################################
# Tests of individual subcommands #############################################
###############################################################################
SIGNOFF_TEST_DATA = [None, 'FAKE_INSTRUMENT']
@pytest.mark.parametrize("instrument", SIGNOFF_TEST_DATA)  # noqa: E302
def test_delete_signoff_subcommand(CLI, instrument):
    """Test delete signoff subcommand"""
    cmd_args = {
        's_id': 'S001122a',
        'signoff_type': 'FAKE_SIGNOFF_TYPE',
    }

    # Generate command
    cmd = "delete signoff {s_id} {signoff_type}".format(**cmd_args)
    if instrument is not None:
        cmd += " {inst}".format(inst=instrument)

    func = 'ligo.gracedb.rest.GraceDb.delete_signoff'
    with mock.patch(func) as mock_cli_func:
        CLI(shlex.split(cmd))

    # Check call count
    assert mock_cli_func.call_count == 1

    # Get args used in function call
    cli_args, cli_kwargs = mock_cli_func.call_args

    # Check args used in function call
    assert cli_args == (cmd_args['s_id'], cmd_args['signoff_type'],)
    assert cli_kwargs == {'instrument': instrument or ''}



from netrc import netrc as _netrc
from netrc import NetrcParseError
import os
import stat

__all__ = ('netrc', 'NetrcParseError')


class netrc(_netrc):  # noqa: N801
    """Subclass of the Python standard library :class:`~netrc.netrc` class to
    add some custom behaviors.

    1.   If the ``NETRC`` environment variable is defined, then use it as
         the default netrc file path.

    2.   Backport permissions checks that were added in Python 3.1
         (see https://bugs.python.org/issue14984).

    3.   Apply permissions checks whether or or not we are reading from the
         default netrc file, and whether or not the file contains passwords.
    """

    def __init__(self, file=None):
        if file is None:
            file = os.environ.get('NETRC')
        _netrc.__init__(self, file)

    def _parse(self, file, fp, *args, **kwargs):
        # Adapted from Python 3.7
        if os.name == 'posix':
            prop = os.fstat(fp.fileno())
            if prop.st_uid != os.getuid():
                import pwd
                try:
                    fowner = pwd.getpwuid(prop.st_uid)[0]
                except KeyError:
                    fowner = 'uid %s' % prop.st_uid
                try:
                    user = pwd.getpwuid(os.getuid())[0]
                except KeyError:
                    user = 'uid %s' % os.getuid()
                raise NetrcParseError(
                    ("~/.netrc file owner (%s) does not match"
                     " current user (%s)") % (fowner, user),
                    file)
            if (prop.st_mode & (stat.S_IRWXG | stat.S_IRWXO)):
                raise NetrcParseError(
                    "~/.netrc access too permissive: access"
                    " permissions must restrict access to only"
                    " the owner", file)
        _netrc._parse(self, file, fp, *args, **kwargs)



#!/usr/bin/env python
# Copyright (C) Brian Moe, Branson Stephens (2015)
#
# This file is part of gracedb
#
# gracedb is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# It is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with gracedb.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import print_function
import os, sys, shutil
import json
import six
from ligo.gracedb.rest import GraceDb
from ligo.gracedb.rest import DEFAULT_SERVICE_URL

DEFAULT_COLUMNS = "graceid,labels,group,pipeline,search,far,gpstime,created,dataurl"

#-----------------------------------------------------------------
# Util routines

def error(*message):
    message = "".join(message) + "\n"
    sys.stderr.write("ERROR: " + message)

def warning(*message):
    message = "".join(message) + "\n"
    sys.stderr.write("WARNING: " + message)

def output(*message):
    message = "".join(message) + "\n"
    sys.stdout.write(message)

def defaultAccess(e,a):
    if a.find('.') < 0:
        return repr(e.get(a,""))
    rv = e
    attrs = a.split('.')
    while attrs and rv:
        rv = rv.get(attrs[0],"")
        attrs = attrs[1:]
    return repr(rv)

#-----------------------------------------------------------------
# HTTP upload encoding
# Taken from http://code.activestate.com/recipes/146306/

# NB:  We are not checking against this list anymore.  Instead, we will
# get the list of groups and types from the API Root.  However, this is
# left in so that the docstring will still work even without a connection
# to the API.
typeCodeMap = {
        "LowMass" : "LM",
        "HighMass" : "HM",
        "GRB"      : "GRB",
        "Ringdown" : "RD",
        "Omega"    : "OM",
        "Q"        : "Q",
        "X"        : "X",
        "CWB"      : "CWB",
        "MBTAOnline": "MBTA",
        "Injection": "HWINJ",
}
validTypes = list(typeCodeMap)

#-----------------------------------------------------------------
# Override and add a few methods to the base GraceDb client class
class GraceDbClient(GraceDb):
    def __init__(self, url=DEFAULT_SERVICE_URL, *args, **kwargs):
        if (url[-1] != '/'):
            url += '/'
        self.url = url
        super(GraceDbClient, self).__init__(url, *args, **kwargs)

    def download(self, graceid, filename, destfile):
        # Check that we *could* write the file before we
        # go to the trouble of getting it.  Also, try not
        # to open a file until we know we have data.
        if not hasattr(destfile, 'read') and destfile != "-":
            if not os.access(os.path.dirname(os.path.abspath(destfile)), os.W_OK):
                raise IOError("%s: Permission denied" % destfile)
        response = self.files(graceid, filename)
        if response.status == 200:
            if not hasattr(destfile, 'read'):
                if destfile == '-':
                    destfile = sys.stdout
                    # Python 2/3 compatibility
                    if hasattr(destfile, 'buffer'):
                        destfile = destfile.buffer
                else:
                    destfile = open(destfile, "wb")
            shutil.copyfileobj(response, destfile)
            return 0
        else:
            return "Error. (%d) %s" % (response.status, response.reason)

    # Hamstring 'adjustResponse' from the example REST client.
    # We don't want it messing with the response from the server.
    def adjustResponse(self, response):
        response.json = lambda: self.load_json_or_die(response)
        return response

    @classmethod
    def output_and_die(cls, msg):
        sys.stderr.write(msg)
        sys.exit(1)

#-----------------------------------------------------------------
# Main


def main(args=None):
    if args is None:
        args = sys.argv[1:]

    usage ="""%%prog [options] GROUP PIPELINE SEARCH EVENTFILE
    where GROUP is one of %(groups)s
         PIPELINE is one of %(pipelines)s
         SEARCH (optional) is one of %(searches)s
         EVENTFILE is file containing event data. '-' indicates stdin.
    NOTE: the groups, pipelines and searches in this docstring may not
    be up top date. To see an accurate list, do:

    %%prog list groups
    %%prog list pipelines
    %%prog list searches

%%prog [options] replace GRACEID EVENTFILE
   where GROUP is one of %(groups)s
         EVENTFILE is file containing event data. '-' indicates stdin.

%%prog [options] ping
   Test server connection

%%prog [options] upload GRACEID FILE [COMMENT]
   where GRACEID is the id of an existing candidate event in GraCEDb
         FILE      is the name of the file to upload. '-' indicates stdin.
         COMMENT   is an optional annotation to enter into the log
   Upload FILE to the private data area for a candidate event. To apply
   a tag, use the --tag-name option (and --tag-display-name if desired.)

%%prog [options] download GRACEID FILE [DESTINATION]
   where GRACEID      is the id of an existing candidate event in GraCEDb
         FILE         is the name of the file previosuly uploaded.
         DESTINATION  is the download destination.  '-' indicates stdout.
                      default is same file name as FILE
    Download FILE from private data area of a candidate event

%%prog [options] log GRACEID COMMENT
   where GRACEID  is the id of an existing candidate event in GraCEDb
         COMMENT  is text that will be entered into the event's log
   Enter a comment into the log for a candidate event.  To apply a tag,
   use the --tag-name option (and --tag-display-name if desired).

%%prog [options] label GRACEID LABEL
    Label event with GRACEDID with LABEL.  LABEL must already exist.

%%prog [options] tag GRACEID LOG_N TAG_NAME [DISP_NAME]
   where GRACEID   is the id of an existing candidate event in GraCEDb
         LOG_N     is the number of the log message.
         TAG_NAME  is the name of the tag
         DISP_NAME is the tag display name (ignored for existing tags)
    Tag an existing log message. Alternatively, the tag name and
    display name can be passed in with the --tag-name and
    --tag-display-name options.

%%prog [options] delete_tag GRACEID LOG_N TAG_NAME
    Remove a tag from a log message. Alternatively, the tag name
    can be passed in with the --tag-name option.

%%prog [options] search SEARCH PARAMS
    Search paramaters are a list of requirements to be satisfied.  They
    may be GPS times, GPS time ranges, graceids and ranges, group(s),
    analysis type(s), labels, etc.  Note that text is case insensitive
    Example: %%prog search G0100..G0200 mbta LUMIN_GO

%%prog version
    Display version information.

Environment Variables:
    GRACEDB_SERVICE_URL   (can be overridden by --service-url)
    X509_USER_PROXY
    X509_USER_CERT
    X509_USER_KEY

Credentials are looked for in this order:
    (1) $(X509_USER_CERT) / $(X509_USER_KEY)
    (2) $(X509_USER_PROXY)
    (3) Default location of grid proxy ( /tmp/x509up_u$(UID) )
    (4) $(HOME)/.globus/usercert.pem / $(HOME)/.globus/userkey.pem
    (5) Basic auth credentials in $(HOME)/.netrc

Note that comments can only be 200 characters long.
Longer strings will be truncated.""" % {
        'groups'     : 'CBC, Burst, Stochastic, Coherent, Test, External',
        'pipelines'  : 'MBTAOnline, gstlal, spiir, oLIB, PyCBC, HardwareInjection, Fermi, Swift, CWB, CWB2G',
        'searches'   : 'AllSky, LowMass, HighMass, GRB, Test',
    }

    from optparse import OptionParser
    op = OptionParser(usage=usage)
    op.add_option("-s", "--service-url", dest="service",
                  help="GraCEDb Service URL", metavar="URL")
    op.add_option("-f", "--filename", dest="filename",
                  help="If data is read from stdin, use this as the filename.", metavar="NAME")
    op.add_option("-c", "--columns", dest="columns",
                  help="Comma separated list of event attributes to include in results (only meaningful in search)",
                  default=DEFAULT_COLUMNS
                 )

    op.add_option("-l", "--ligolw", dest="ligolw",
                  help="Download ligolw file of combined search results (not meaningful outside of search). NOTE: Produces an ERROR if any of the events returned by the search do not have coinc.xml files.",
                  action="store_true", default=False
                 )
    op.add_option("-t", "--tag-name", dest="tagName",
                  help="tag name in database (only used for log, upload, tag, and delete_tag)",
                  default=None
                 )
    op.add_option("-d", "--tag-display-name", dest="tagDispName",
                  help="tag display name (ignored for existing tags)",
                  default=None
                 )
    op.add_option("-n", "--force-noauth", dest="force_noauth",
                  help="Do not use any authentication credentials.",
                  action="store_true", default=False
                 )
    op.add_option("-o", "--offline", dest="offline", action="store_true",
                  help=("Signifies that an event was found by an offline "
                  "pipeline. Used when creating a new event."), default=False
                 )
    op.add_option("-a", "--username", dest="username", type=str,
                  help=("Username for basic auth"), default=None
                 )
    op.add_option("-b", "--password", dest="password", type=str,
                  help=("Password for basic auth"), default=None
                 )
    op.add_option("--labels", dest="labels",help=("Defines labels for creating"
                  " a new event. Should be a comma-separated list of labels."),
                  default=[], type=str
                 )
    op.add_option("--show-creds", dest="show_creds", help=("Prints your "
                  "authentication credentials and exits."), default=False,
                  action="store_true"
                 )

    options, args = op.parse_args(args)
    if isinstance(options.labels, str):
        options.labels = options.labels.split(',')

    try:
        from glue.ligolw import ligolw
        from glue.ligolw import lsctables
        from glue.ligolw import utils
        from glue.ligolw.utils import ligolw_add

        lsctables.use_in(ligolw.LIGOLWContentHandler)
    except:
        if options.ligolw:
            error("ligolw modules not found")
            exit(1)
        else:
            pass

    # warn users about depreciation
    warning("The gracedb_legacy CLI has been deprecated and "
        "will be removed in a future release. Please use the updated "
        "gracedb CLI or available API tools for your workflow.")

    service = options.service or \
              os.environ.get('GRACEDB_SERVICE_URL', None) or \
              DEFAULT_SERVICE_URL

    # Compile dict of kwargs for instantiating the client class
    client_args = {}

    # Auth args
    client_args['force_noauth'] = options.force_noauth
    client_args['username'] = options.username
    client_args['password'] = options.password

    # Instantiate the client
    client = GraceDbClient(service, **client_args)

    if options.show_creds:
        client.show_credentials()
        print()
        exit(0)

    if len(args) < 1:
        op.error("not enough arguments")
    elif args[0] == 'version':
        import pkg_resources
        version = pkg_resources.require("ligo-gracedb")[0].version
        print("GraceDB Client v. %s" % version)
        exit(0)
    elif args[0] == 'ping':
        response = client.ping()
        output("Client groups: %s" % client.groups)
        output("Client pipelines: %s" % client.pipelines)
        output("Client searches: %s" % client.searches)
        if response.status==200:
            output("%s: 200 OK" % service)
            exit(0)
    elif args[0] == 'list':
        if args[1] == 'groups':
            output(' '.join(client.groups))
            exit(0)
        elif args[1] == 'pipelines':
            output(' '.join(client.pipelines))
            exit(0)
        elif args[1] == 'searches':
            output(' '.join(client.searches))
            exit(0)
        else:
            output("Unknown list object. Please use 'groups', 'pipelines', or 'searches.'")
            exit(1)
    elif args[0] == 'upload':
        if len(args) < 3:
            op.error("not enough arguments for upload")
        graceid = args[1]
        filename = args[2]
        comment = " ".join(args[3:])
        tagName = options.tagName
        tagDispName = options.tagDispName
        response = client.writeLog(graceid, comment, filename, None,
            tagName, tagDispName)
    elif args[0] == 'download':
        if len(args) not in [2,3,4]:
            op.error("not enough arguments for download")
        graceid = args[1]
        if len(args) == 2:
            # get/print listing.
            response = client.files(graceid)
            if response and response.status == 200:
                for fname in json.loads(response.read()):
                    print(fname)
                exit(0)
            print(response.reason)
            exit(1)
        filename = args[2]
        if len(args) == 4:
            outfile = args[3]
        else:
            outfile = os.path.basename(filename)
        response = client.download(graceid, filename, outfile)
        if response:
            # no response means file saved.  any other response is an error message.
            print(response)
            exit(1)
        exit(0)
    elif args[0] == 'log':
        if len(args) < 3:
            op.error("not enough arguments for log")
        graceid = args[1]
        message = " ".join(args[2:])
        response = client.writeLog(graceid, message, tagname=options.tagName, displayName=options.tagDispName)
    elif args[0] == 'tag':
        if options.tagName:
            if len(args) != 2:
                op.error("wrong number of arguments for tag")
            tagName = options.tagName
            tagDispName = options.tagDispName
        else:
            if len(args) not in [4,5]:
                op.error("wrong number of arguments for tag")
            tagName = args[3]
            tagDispName = None
            if len(args)==5:
                tagDispName = args[4]
        graceid = args[1]
        logN = args[2]
        response = client.addTag(graceid, logN, tagName, tagDispName)
    elif args[0] == 'delete_tag':
        error("len args = %s" % len(args))
        error("args = %s" % args)
        if options.tagName:
            if len(args) != 2:
                op.error("wrong number of arguments for delete_tag")
            tagName = options.tagName
        else:
            if len(args) != 4:
                op.error("wrong number of arguments for delete_tag")
            tagName = args[3]
        graceid = args[1]
        logN = args[2]
        response = client.removeTag(graceid, logN, tagName)
    elif args[0] == 'label':
        if len(args) != 3:
            op.error("wrong number of arguments for label")
        graceid = args[1]
        label = args[2]
        response = client.writeLabel(graceid, label)
    elif args[0] == 'search':
        query = " ".join(args[1:])

        columns = options.columns
        columns = columns.replace('DEFAULTS',DEFAULT_COLUMNS)
        columns = columns.split(',')

        count = None # XXX Let's just get rid of this?
        orderby = None # XXX Should we implement this?

        events = client.events(query, orderby, count, columns)

        if options.ligolw:
            xmldoc = ligolw.Document()
            for e in events:
                graceid = e['graceid']
                try:
                    r = client.files(graceid, "coinc.xml")
                    utils.load_fileobj(r, xmldoc = xmldoc, contenthandler = ligolw.LIGOLWContentHandler)
                except:
                    error("Missing coinc.xml for %s. Cannot build ligolw output." % graceid)
                    exit(1)
            ligolw_add.reassign_ids(xmldoc)
            ligolw_add.merge_ligolws(xmldoc)
            ligolw_add.merge_compatible_tables(xmldoc)
            xmldoc.write()
        else:
            accessFun = {
                "labels" : lambda e: \
                    ",".join(list(e['labels'])),
                "dataurl" : lambda e: e['links']['files'],
            }

            header = "#" + "\t".join(columns)
            output(header)
            for e in events:
                row = [ accessFun.get(column, lambda e: defaultAccess(e,column))(e) for column in columns ]
                row = "\t".join(row)
                output(row)

        return 0
    elif args[0] == 'replace':
        if len(args) != 3:
            op.error("wrong number of args for replace")
        graceid = args[1]
        filename = args[2]
        response = client.replaceEvent(graceid, filename)
    elif len(args) in [3,4]:
        # Create a new event.
        group = args[0]
        pipeline = args[1]
        if len(args)==3:
            search = None
            filename = args[2]
        else:
            search = args[2]
            filename = args[3]

        # Check that the group, search, and pipeline are known to the API.
        foundGroup = True if (six.text_type(group) in client.groups) else False
        if not foundGroup:
            error("Group must be one of: ", ", ".join(client.groups))
            sys.exit(1)

        foundPipeline = True if (six.text_type(pipeline) in client.pipelines) else False
        if not foundPipeline:
            error("Pipeline must be one of: ", ", ".join(client.pipelines))
            sys.exit(1)

        if search:
            foundSearch = True if (six.text_type(search) in client.searches) else False
            if not foundSearch:
                error("Search must be one of: ", ", ".join(client.searches))
                sys.exit(1)

        response = client.createEvent(group, pipeline, filename, search,
            offline=options.offline, labels=options.labels)

        if not response:
            error("There was a problem.  Did you do ligo-proxy-init?")
            sys.exit(1)

        # XXX Must output graceid for consistency with earlier client.
        # Therefore, must deal with response here rather than at the end.
        exitCode = 0
        status = response.status
        if status >= 400:
            exitCode=1
        try:
            rv = response.read()
        except:
            rv = response
        try:
            rv = json.loads(rv.decode())
        except:
            pass

        if 'graceid' in list(rv):
            output(rv['graceid'])
        elif 'error' in list(rv):
            exitCode=1
            error(rv['error'])

        return exitCode

    else:
        op.error("")
        sys.exit(1)

    # Output the response.
    exitCode = 0
    try:
        rv = response.read()
        status = response.status
    except:
        rv = response

    try:
        responseBody = json.loads(rv.decode())
    except:
        responseBody = rv.decode()

    if status >= 400:
        exitCode=1
    if isinstance(responseBody, str):
        output("%d: %s" % (status, responseBody))
    else:
        output("Server returned %d" % status)
        if ('error' in responseBody) and response['error']:
            error(response['error'])
            exitCode = 1
        if ('warning' in responseBody) and response['warning']:
            warning(response['warning'])
        if ('output' in responseBody) and response['output']:
            output(response['output'])

    return exitCode

if __name__ == "__main__":
    code = main()
    sys.exit(code)



# Tests for subcommands below 'create'
#  Ex: 'gracedb create emobservation'
import datetime
import pytest
import shlex
import six
try:
    from unittest import mock
except ImportError:
    import mock

# Apply module-level mark
pytestmark = pytest.mark.cli


###############################################################################
# Tests of individual subcommands #############################################
###############################################################################
EMO_TEST_DATA = [
    ([0.1, 0.2, 0.9, 0.7], [0.1, 0.2, 0.04, 1.2], [1, 2, 3, 4], None),
    (0.2, 0.1, 2.2, "test comment"),
]
@pytest.mark.parametrize("ra_width,dec_width,duration,comment",  # noqa: E302
                         EMO_TEST_DATA)
def test_create_emobservation_subcommand(CLI, ra_width, dec_width, duration,
                                         comment):
    """Test create observation subcommand"""
    now = datetime.datetime.now()
    s_id = 'S001122a'
    group = 'FAKE_EMGROUP'
    ra_list = [1, 2, 3, 4]
    dec_list = [5, 6, 7, 8]
    start_time_list = list(
        map(lambda i: (now + datetime.timedelta(seconds=i)).isoformat(),
            [0, 1, 2, 3])
    )

    # Compile command
    def join_s(s):
        return ",".join([str(i) for i in s]) if isinstance(s, list) else s

    cmd = ('create emobservation {s_id} {group} {ra_list} {ra_width_list} '
           '{dec_list} {dec_width_list} {start_time_list} {duration_list}') \
        .format(s_id=s_id, group=group, ra_list=join_s(ra_list),
                ra_width_list=join_s(ra_width), dec_list=join_s(dec_list),
                dec_width_list=join_s(dec_width),
                duration_list=join_s(duration),
                start_time_list=join_s(start_time_list))
    if comment is not None:
        cmd += " --comment='{comment}'".format(comment=comment)

    func = 'ligo.gracedb.rest.GraceDb.writeEMObservation'
    with mock.patch(func) as mock_cli_func:
        CLI(shlex.split(cmd))

    # Check call count
    assert mock_cli_func.call_count == 1

    # Get args used in function call
    cli_args, cli_kwargs = mock_cli_func.call_args

    # Check args used in function call
    assert cli_args == (s_id, group, ra_list, ra_width, dec_list, dec_width,
                        start_time_list, duration,)
    assert cli_kwargs == {'comment': comment}


EVENT_TEST_DATA = [
    (None, None, None),
    (None, 'FAKE_LABEL', True),
    ('FAKE_SEARCH', ['FAKE_LABEL1', 'FAKE_LABEL2'], False),
]
@pytest.mark.parametrize("search,labels,offline",  # noqa: #302
                         EVENT_TEST_DATA)
def test_create_event_subcommand(CLI, search, labels, offline):
    """Test create event subcommand"""
    cmd_args = {
        'group': 'FAKE_GROUP',
        'pipeline': 'FAKE_PIPELINE',
        'event_file': '/path/to/fake/file.xml',
    }

    # Generate command
    cmd = 'create event {group} {pipeline} {event_file}'.format(**cmd_args)
    if search is not None:
        cmd += ' {search}'.format(search=search)
    if labels is not None:
        if isinstance(labels, list):
            cmd += " --labels={labels}".format(labels=",".join(labels))
        else:
            cmd += " --labels={labels}".format(labels=labels)
    if offline:
        cmd += " --offline"

    func = 'ligo.gracedb.rest.GraceDb.createEvent'
    func2 = 'ligo.gracedb.rest.GraceDb.searches'
    with mock.patch(func2, new_callable=mock.PropertyMock) as mock_searches, \
         mock.patch(func) as mock_cli_func:  # noqa: E127

        # Set up return value and call CLI
        mock_searches.return_value = []
        CLI(shlex.split(cmd))

    # Check call count
    assert mock_cli_func.call_count == 1

    # Get args used in function call
    cli_args, cli_kwargs = mock_cli_func.call_args

    # Check args used in function call
    assert cli_args == (cmd_args['group'], cmd_args['pipeline'],
                        cmd_args['event_file'],)
    assert cli_kwargs == {'search': search, 'labels': labels,
                          'offline': offline or False}


def test_create_event_legacy_usage(CLI):
    """Test create event legacy usage (search and event file are swapped)"""
    cmd_args = {
        'group': 'FAKE_GROUP',
        'pipeline': 'FAKE_PIPELINE',
        'event_file': '/path/to/fake/file.xml',
        'search': 'FAKE_SEARCH',
    }

    # Generate command
    cmd = 'create event {group} {pipeline} {search} {event_file}'.format(
        **cmd_args)

    # Have to patch client.searches and os.path.isfile
    func = 'ligo.gracedb.rest.GraceDb.createEvent'
    func2 = 'ligo.gracedb.rest.GraceDb.searches'
    func3 = 'ligo.gracedb.cli.commands.create.os.path.isfile'
    with mock.patch(func2, new_callable=mock.PropertyMock) as mock_searches, \
         mock.patch(func3) as mock_isfile, \
         mock.patch(func) as mock_cli_func:  # noqa: E127

        # Set up return values
        mock_searches.return_value = cmd_args['search']
        mock_isfile.return_value = True

        # Call CLI
        CLI(shlex.split(cmd))

    # Check call count
    assert mock_cli_func.call_count == 1

    # Get args used in function call
    cli_args, cli_kwargs = mock_cli_func.call_args

    # Check args used in function call
    assert cli_args == (cmd_args['group'], cmd_args['pipeline'],
                        cmd_args['event_file'],)
    assert cli_kwargs == {'search': cmd_args['search'], 'labels': None,
                          'offline': False}


LOG_TEST_DATA = [
    (None, None, None),
    (None, 'FAKE_TAG', 'FAKE_DISP_NAME'),
    ('/path/to/fake/file.txt', ['FAKE_TAG1', 'FAKE_TAG2'],
        ['FAKE_DISP_NAME1', 'FAKE_DISP_NAME2']),
]
@pytest.mark.parametrize("filename,tag_name,tag_disp_name",  # noqa: E302
                         LOG_TEST_DATA)
def test_create_log_subcommand(CLI, filename, tag_name, tag_disp_name):
    """Test create log subcommand"""
    cmd_args = {
        's_id': 'S001122a',
        'comment': 'test comment',
    }

    # Generate command
    cmd = "create log {s_id} '{comment}'".format(**cmd_args)
    if filename is not None:
        cmd += " {filename}".format(filename=filename)
    if tag_name is not None:
        tag_name_str = tag_name
        if isinstance(tag_name_str, list):
            tag_name_str = ",".join(tag_name_str)
        cmd += " --tag-name={tag_name}".format(tag_name=tag_name_str)
    if tag_disp_name is not None:
        tag_disp_str = tag_disp_name
        if isinstance(tag_disp_str, list):
            tag_disp_str = ",".join(tag_disp_str)
        cmd += " --tag-display-name={tag_disp_name}".format(
            tag_disp_name=tag_disp_str)

    func = 'ligo.gracedb.rest.GraceDb.writeLog'
    with mock.patch(func) as mock_cli_func:
        CLI(shlex.split(cmd))

    # Check call count
    assert mock_cli_func.call_count == 1

    # Get args used in function call
    cli_args, cli_kwargs = mock_cli_func.call_args

    # Check args used in function call
    assert cli_args == (cmd_args['s_id'], cmd_args['comment'],)
    assert cli_kwargs == {'filename': filename, 'tag_name': tag_name,
                          'displayName': tag_disp_name}


SIGNOFF_TEST_DATA = [None, 'FAKE_INSTRUMENT']
@pytest.mark.parametrize("instrument", SIGNOFF_TEST_DATA)  # noqa: E302
def test_create_signoff_subcommand(CLI, instrument):
    """Test create signoff subcommand"""
    cmd_args = {
        's_id': 'S001122a',
        'signoff_type': 'FAKE_SIGNOFF_TYPE',
        'signoff_status': 'FAKE_SIGNOFF_STATUS',
        'comment': 'test comment',
    }

    # Generate command
    cmd = "create signoff {s_id} {signoff_type} {signoff_status} '{comment}'" \
        .format(**cmd_args)
    if instrument is not None:
        cmd += " {inst}".format(inst=instrument)

    func = 'ligo.gracedb.rest.GraceDb.create_signoff'
    with mock.patch(func) as mock_cli_func:
        CLI(shlex.split(cmd))

    # Check call count
    assert mock_cli_func.call_count == 1

    # Get args used in function call
    cli_args, cli_kwargs = mock_cli_func.call_args

    # Check args used in function call
    assert cli_args == (cmd_args['s_id'], cmd_args['signoff_type'],
                        cmd_args['signoff_status'], cmd_args['comment'],)
    assert cli_kwargs == {'instrument': instrument or ''}


SUPEREVENT_TEST_DATA = [
    (None, None, ['G1234', 'G1235']),
    (None, 'FAKE_LABEL', 'G1236'),
    ('FAKE_CATEGORY', ['FAKE_LABEL1', 'FAKE_LABEL2'], None),
]
@pytest.mark.parametrize("category,labels,events",  # noqa: E302
                         SUPEREVENT_TEST_DATA)
def test_create_superevent_subcommand(CLI, category, labels, events):
    """Test create superevent subcommand"""
    cmd_args = {
        't_start': 1,
        't_0': 2,
        't_end': 3,
        'pref_ev': 'G0001',
    }

    # Generate command
    cmd = "create superevent {t_start} {t_0} {t_end} {pref_ev}".format(
        **cmd_args)
    if category is not None:
        cmd += " --category={cat}".format(cat=category)
    if labels is not None:
        if isinstance(labels, list):
            cmd += " --labels={labels}".format(labels=",".join(labels))
        else:
            cmd += " --labels={labels}".format(labels=labels)
    if events is not None:
        if isinstance(events, list):
            cmd += " --events={events}".format(events=",".join(events))
        else:
            cmd += " --events={events}".format(events=events)

    func = 'ligo.gracedb.rest.GraceDb.createSuperevent'
    with mock.patch(func) as mock_cli_func:
        CLI(shlex.split(cmd))

    # Check call count
    assert mock_cli_func.call_count == 1

    # Get args used in function call
    cli_args, cli_kwargs = mock_cli_func.call_args

    # Check args used in function call
    assert cli_args == (cmd_args['t_start'], cmd_args['t_0'],
                        cmd_args['t_end'], cmd_args['pref_ev'],)
    assert cli_kwargs == {'category': category or 'production',
                          'labels': labels, 'events': events}


# Define a dict where keys are parameters and values are lists of
# values for those parameters
VOEVENT_TEST_DICT = {
    'skymap_type': [None, 'FAKE_SKYMAP_TYPE'],
    'skymap_filename': [None, 'FAKE_SKYMAP_FILENAME'],
    'external': [None, True],
    'open_alert': [None, True],
    'hardware_inj': [None, True],
    'coinc_comment': [None, True],
    'prob_has_ns': [None, 0.1],
    'prob_has_remnant': [None, 0.2],
    'bns': [None, 0.3],
    'nsbh': [None, 0.4],
    'bbh': [None, 0.5],
    'terrestrial': [None, 0.6],
    'mass_gap': [None, 0.7],
}
# Convert into a list of tuples
VOEVENT_TEST_DATA = zip(*list(six.itervalues(VOEVENT_TEST_DICT)))
@pytest.mark.parametrize(",".join(list(VOEVENT_TEST_DICT)),  # noqa: E302
                         VOEVENT_TEST_DATA)
def test_create_voevent_subcommand(
    CLI, skymap_type, skymap_filename, external, open_alert, hardware_inj,
    coinc_comment, prob_has_ns, prob_has_remnant, bns, nsbh, bbh, terrestrial,
    mass_gap
):
    """Test create voevent subcommand"""
    cmd_args = {
        's_id': 'S001122a',
        'voevent_type': 'FAKE_VOEVENT_TYPE',
    }

    # Generate command
    cmd = "create voevent {s_id} {voevent_type}".format(**cmd_args)
    if skymap_type is not None:
        cmd += " --skymap-type={st}".format(st=skymap_type)
    if skymap_filename is not None:
        cmd += " --skymap-filename={sf}".format(sf=skymap_filename)
    if external:
        cmd += " --external"
    if open_alert:
        cmd += " --open-alert"
    if hardware_inj:
        cmd += " --hardware-inj"
    if coinc_comment:
        cmd += " --coinc-comment"
    if prob_has_ns is not None:
        cmd += " --prob-has-ns={phn}".format(phn=prob_has_ns)
    if prob_has_remnant is not None:
        cmd += " --prob-has-remnant={phr}".format(phr=prob_has_remnant)
    if bns is not None:
        cmd += " --bns={bns}".format(bns=bns)
    if nsbh is not None:
        cmd += " --nsbh={nsbh}".format(nsbh=nsbh)
    if bbh is not None:
        cmd += " --bbh={bbh}".format(bbh=bbh)
    if terrestrial is not None:
        cmd += " --terrestrial={terr}".format(terr=terrestrial)
    if mass_gap is not None:
        cmd += " --mass-gap={mg}".format(mg=mass_gap)

    func = 'ligo.gracedb.rest.GraceDb.createVOEvent'
    with mock.patch(func) as mock_cli_func:
        CLI(shlex.split(cmd))

    # Check call count
    assert mock_cli_func.call_count == 1

    # Get args used in function call
    cli_args, cli_kwargs = mock_cli_func.call_args

    # Check args used in function call
    assert cli_args == (cmd_args['s_id'], cmd_args['voevent_type'],)
    assert cli_kwargs == {
        'skymap_type': skymap_type,
        'skymap_filename': skymap_filename,
        'internal': not (external or False),
        'open_alert': open_alert or False,
        'hardware_inj': hardware_inj or False,
        'CoincComment': coinc_comment or False,
        'ProbHasNS': prob_has_ns,
        'ProbHasRemnant': prob_has_remnant,
        'BNS': bns,
        'NSBH': nsbh,
        'BBH': bbh,
        'Terrestrial': terrestrial,
        'MassGap': mass_gap,
    }



# -*- coding: utf-8 -*-
# Copyright (C) Alexander Pace, Tanner Prestegard,
#               Branson Stephens, Brian Moe (2020)
#
# This file is part of gracedb
#
# gracedb is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# It is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with gracedb.  If not, see <http://www.gnu.org/licenses/>.
import os
import sys
import json as json_lib

from warnings import warn
from requests import Session
from .extern.safe_netrc import netrc
from os import getuid
from .version import __version__
from .adapter import GraceDbCertAdapter
from .utils import hook_response, raise_status_exception

# To remove later: python2 compatibility fix:
if sys.version_info[0] > 2:
    from urllib.parse import urlparse
else:
    from urlparse import urlparse

DEFAULT_SERVICE_URL = "https://gracedb.ligo.org/api/"


class GraceDBClient(Session):
    """
    url (:obj:`str`, optional): URL of server API
    cred (:obj:`tuple` or :obj:`str, optional): a tuple or list of
        (``/path/to/cert/file``, ``/path/to/key/file) or a single path to
        a combined proxy file (if using an X.509 certificate for
        authentication)
    username (:obj:`str`, optional): username for basic auth
    password (:obj:`str`, optional): password for basic auth
    force_noauth (:obj:`bool`, optional): set to True if you want to skip
        credential lookup and use this client as an unauthenticated user
    fail_if_noauth (:obj:`bool`, optional): set to True if you want the
        constructor to fail if no authentication credentials are provided
        or found
    reload_certificate (:obj:`bool`, optional): if ``True``, your
        certificate will be checked before each request whether it is
        within ``reload_buffer`` seconds of expiration, and if so, it will
        be reloaded. Useful for processes which may live longer than the
        certificate lifetime and have an automated method for certificate
        renewal. The path to the new/renewed certificate **must** be the
        same as for the old certificate.
    reload_buffer (:obj:`int`, optional): buffer (in seconds) for reloading
        a certificate in advance of its expiration. Only used if
        ``reload_certificate`` is ``True``.

    Authentication details:
    You can:
        1. Provide a path to an X.509 certificate and key or a single
           combined proxy file
        2. Provide a username and password
    Or:
        The code will look for a certificate in a default location
            (``/tmp/x509up_u%d``, where ``%d`` is your user ID)
        The code will look for a username and password for the specified
            server in ``$HOME/.netrc``
    """

    def __init__(self, url=DEFAULT_SERVICE_URL, cred=None, username=None,
                 password=None, force_noauth=False, fail_if_noauth=False,
                 reload_certificate=False, reload_buffer=300,
                 *args, **kwargs):
        super(GraceDBClient, self).__init__(*args, **kwargs)

        # Initialize variables:
        self.cert = None
        self.auth = None
        self.host = urlparse(url).hostname
        self.auth_type = None

        # Set up credentials. First, only attempt if the user did not
        # force noauth:
        if not force_noauth:
            # Attempt to assign x509 credentials.
            if cred or not (username or password):
                self.cert = self._get_x509_credentials(cred)

            # if we weren't able to determine a x509 certificate,
            # then look for basic auth.
            if not self.cert:
                self.auth = self._process_basic_credentials(username,
                                                            password)

                # If no basic auth credentials were provided, and
                # fail_if_noauthis set, then fail.
                if not self.auth:
                    if fail_if_noauth:
                        raise RuntimeError("No authentication credentials "
                                           "could be found, and fail_if_"
                                           "noauth is set.")
                    else:
                        warn("Authentication credentials not found, "
                             "proceeding with unauthorized session")

        elif fail_if_noauth:
            raise ValueError('You have provided conflicting parameters '
                             'to the client constructor: '
                             'fail_if_noauth=True and force_noauth=True.')

        # Update session headers:
        self.headers.update(self._update_headers())

        # Adjust the response via a session hook:
        self.hooks = {'response': [hook_response, raise_status_exception]}

        if reload_certificate and self.auth_type == 'x509':
            self.mount('https://', GraceDbCertAdapter(
                       cert=self.cert,
                       reload_buffer=reload_buffer))

    def _process_basic_credentials(self, username, password):
        """ Gathers basic auth credentials. First it looks for
            the input username/password, then checks the netrc file
        """

        # Checks username/password:
        if username and password:
            self.auth_type = 'basic'
            return username, password
        # If one or the other is provided, but not both, raise an error.
        elif bool(username) != bool(password):
            raise RuntimeError('You must provide both a username and a '
                               'password for basic authentication.')
        # Finally, try the netrc fle:
        else:
            try:
                netrc_auth = netrc().authenticators(self.host)
            except IOError:
                pass
            else:
                if netrc_auth is not None:
                    self.auth_type = 'basic'
                    return netrc_auth[0], netrc_auth[2]

        # If that didn't work, then give up.
        return

    def _get_x509_credentials(self, cred):
        """ Get a user's x509 credentials. Look in the following locations.
               1) 'cred' input variable, either cert/key pair or combined file
               2) $X509_USER_PROXY/CERT/KEY environment variables.
                  - X509_USER_PROXY: combined cert/cert pair
                  - X509_USER_CERT: x509 certificate file.
                  - X509_USER_KEY: corresponding x509 private key.
               3) Auto-generated grid proxy in /tmp/x509up_$UID
        """

        # First use the contents of the 'cred' variable.
        # Assume that you'll find the x509 credentials *somewhere*
        # in here. Switch it back at the end if you don't.
        self.auth_type = 'x509'
        if cred:
            # if it's a list or a tuple, then return the pair
            # as a tuple to the session
            if isinstance(cred, (list, tuple)):
                return tuple(cred)
            elif isinstance(cred, str) and os.path.exists(cred):
                return cred
            else:
                warn("x509 cert/key/pair provided by 'cert' variable "
                     "is invalid, or not found.")
                return None

        # Now: check envionment variables:
        grid_proxy_file = os.environ.get('X509_USER_PROXY')
        grid_cert_file = os.environ.get('X509_USER_CERT')
        grid_key_file = os.environ.get('X509_USER_KEY')

        # Set logic for returning credentials:
        # If the user supplies a proxy file, then return it:
        if grid_proxy_file:
            return grid_proxy_file
        # If the user supplies both a cert and key, return those:
        elif grid_cert_file and grid_key_file:
            return grid_cert_file, grid_key_file
        # If the user supplies one or the other, then return a warning
        elif grid_cert_file or grid_key_file:
            warn("Warning: must supply a $X509_USER_PROXY or a "
                 "X509_USER_CERT/KEY pair. Proceeding with unauthenticated "
                 "session")
            return None

        # Next try the usual grid-proxy location:
        uid = getuid()
        if uid is not None:
            grid_proxy_file = os.path.join('/tmp', 'x509up_u{}'.format(uid))
            if os.path.exists(grid_proxy_file):
                return grid_proxy_file

        # Finally the default globus key/cert files:
        homedir = os.environ.get('HOME')
        if homedir:
            globus_cert_file = os.path.join(homedir, '.globus', 'usercert.pem')
            globus_key_file = os.path.join(homedir, '.globus', 'userkey.pem')

            if (os.path.exists(globus_cert_file)
                and os.path.exists(globus_key_file)):
                return globus_cert_file, globus_key_file

        # Give up otherwise:
        warn("No x509 credentials found")
        self.auth_type = None

        return None

    def _update_headers(self):
        """ Update the sessions' headers:
        """
        new_headers = {}
        # Assign the user agent. This shows up in gracedb's log files.
        new_headers.update({'User-Agent':
                            'gracedb-client/{}'.format(__version__)})
        new_headers.update({'Accept-Encoding':
                            'gzip, deflate'})
        return new_headers

    # hijack 'Session.request':
    # https://2.python-requests.org/en/master/api/#requests.Session.request
    def request(
            self, method, url, params=None, data=None, headers=None,
            cookies=None, files=None, auth=None, timeout=None,
            allow_redirects=True, proxies=None, hooks=None, stream=None,
            verify=None, cert=None, json=None):
        return super(GraceDBClient, self).request(
            method, url, params=params, data=data, headers=headers,
            cookies=cookies, files=files, auth=auth,
            timeout=timeout, allow_redirects=True, proxies=proxies,
            hooks=hooks, stream=stream, verify=verify, cert=cert, json=json)

    # Extra definitions to return closed contexts' connections
    # back to the pool:
    # https://stackoverflow.com/questions/48160728/resourcewarning
    # -unclosed-socket-in-python-3-unit-test
    def close(self):
        self.close()

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.close()

    # For getting files, return a "raw" file-type object.
    # Automatically decode content.
    def get_file(self, url, **kwargs):
        resp = self.get(url, stream=True, **kwargs)
        resp.raw.decode_content = True
        resp.raw.status_code = resp.status_code
        return resp.raw

    # Return client credentials
    def show_credentials(self, print_output=True):
        """ Prints authentication type and credentials info."""
        output = {'auth_type': self.auth_type}

        cred = {}

        if self.auth_type == 'x509':
            if isinstance(self.cert, tuple):
                cred = {'cert_file': self.cert[0],
                        'key_file': self.cert[1]}
            elif isinstance(self.cert, str):
                cred = {'cert_file': self.cert,
                        'key_file': self.cert}
            else:
                raise ValueError("Problem reading authentication certificate")

        if self.auth_type == 'basic':
            cred = {'username': self.auth[0],
                    'password': self.auth[1]}

        output.update(cred)

        if print_output:
            print(output)
        else:
            return output

    def get_user_info(self):
        """Get information from the server about your user account."""
        user_info_link = self.links.get('user-info', None)
        if user_info_link is None:
            raise RuntimeError('Server does not provide a user info endpoint')
        return self.get(user_info_link)

    @classmethod
    def load_json_from_response(cls, response):
        """ Always return a json content response, even when the server
            provides a 204: no content"""
        # Check if response exists:
        if not response:
            raise ValueError("No response object provided")

        # Check if there is response content. If not, create it.
        if response.content == 'No Content':
            response_content = '{}'

        # Some responses send back strings of strings. This iterates
        # until proper dict is returned, or if it doesn't make progress.
        num_tries = 1
        response_content = response.content.decode('utf-8')

        while type(response_content) == str and num_tries < 3:
            response_content = json_lib.loads(response_content)
            num_tries += 1

        if type(response_content) == dict:
            return response_content
        else:
            return ValueError("ERROR: got unexpected content from "
                              "the server: {}".format(response_content))
